/*
 * Octopus Load Balancer - Balancing algorithms and server connections
 *
 * Copyright 2008 Alistair Reay <alreay1@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 */

/* verifies that the selected member and clone are truly valid.
returns 0 if both member and clone are valid, or if member is valid and cloning is not enabled
returns 1 if member is valid but clone is not
returns -1 if member is invalid
*/
int verify_selection() {
    if(group->overload_mode == MODE_STRICT) {
        if( (group->members[next_member].status != SERVER_STATE_ENABLED) || (group->members[next_member].e_load > 100) || (group->members[next_member].c >= group->members[next_member].maxc) ) {
            return -1;
        }
		if(group->clone_mode != MODE_OFF) {
	        if( (group->clones[next_clone].status != SERVER_STATE_ENABLED) || (group->clones[next_clone].e_load > 100) || (group->clones[next_clone].c >= group->clones[next_clone].maxc) ) {
    	        return 1;
        	}
		}
    }
    else {
        if( (group->members[next_member].status != SERVER_STATE_ENABLED) || (group->members[next_member].c >= group->members[next_member].maxc) ) {
            return -1;
        }
		if(group->clone_mode != MODE_OFF) {
        	if( (group->clones[next_clone].status <= 0) || (group->clones[next_clone].c >= group->clones[next_clone].maxc) ) {
            	return 1;
	        }
		}
    }
    return 0;

}

/* turns a request URI into an integer */
unsigned short int DJBHash(char* str, unsigned int len) {
	unsigned int hash = 5381;
	unsigned int i    = 0;

	for(i = 0; i < len; str++, i++)	{
		hash = ((hash << 5) + hash) + (*str);
	}
	return (unsigned short int)(hash);
}

int choose_hash_server(SESSION *session) {
	if(group->nmembers == 0) {
		return -1;
	}
	int status;
	status=set_hash_server(session);
	if(status < 0) {
		return status;
	}
	status = verify_selection();
	if(status < 0) {
		return status;
	}
	if(status == 1) {
		status=connect_server(session, NO_USE_CLONE);
		return status;
	}
	status=connect_server(session, USE_CLONE);
	return status;

}
/* hashes the request URI to an integer between 0 and group->nmembers and then stores it so the next time the uri is requested it goes back to the same server. this algorithm is designed for HTTP connections only */
int set_hash_server(SESSION *session) {

	int status;
	unsigned short int hash;
	char request_line[URI_LINE_LEN];
	memset(request_line, '\0', URI_LINE_LEN);
	char *uri_start = NULL;
	char *uri = NULL;

	/* copy up to URI_LINE_LEN characters from the first line of the request */
	memccpy(request_line, session->client_read_buffer, '\n', URI_LINE_LEN);
	/* get a pointer to the second word of the request */
	uri_start=strchr(request_line, ' ');
	if(uri_start != NULL) {
		/* then get that second word of the request (this is the uri) */
		uri = strtok(uri_start," ");
		/* strip off query term when hashing uri */
		if(strchr(uri, '?') != NULL) {
			char *result = NULL;
			/* and then move the requested URL (minus the query string) into a temporary buffer */
			result = strtok(uri, uri_delim);
			/* hash the stripped URI to get a number between 0 and HASHTABLESIZE */
			hash=DJBHash(result, (unsigned int)strlen(result));
			hash= hash % (HASHTABLESIZE -1);
		}
		else {
			/* if there is no query term, hash the full URI to get a number between 0 and HASHTABLESIZE */
			hash=DJBHash(uri, (unsigned int)strlen(uri));
			hash= hash % (HASHTABLESIZE -1);
		}

	}
	/* if the strchr function returns NULL it is because the request is not a valid http request, use LC method */
	else {
		set_lc_server();
		return 0;
	}

	/* HASH HIT: if this hash has a server associated with it, set member */
	if (group->hash_table[hash] != HASH_VAL_FREE) {
		if (debug_level > 1) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: algorithm HASH: hash hit for uri %s", uri);
			write_log(OCTOPUS_LOG_STD, log_string);
		}
		next_member=group->hash_table[hash];
		if ((group->clone_mode > MODE_OFF) && (group->nclones > 0)) {
			next_clone=next_member % group->nclones;
		}
		/* check that the server we will use is alive and not overloaded */
		status=verify_selection();
		if(status == 0) {
			return 0;
		}
		/* clone is not available. get new next_member and next_clone vals then set next_member back to how it should be. shortcoming of selection algorithm setting both clone and member no matter what */
		if(status == 1) {
			if(group->clone_mode <= MODE_OFF) {
				return 0;
			}
			else {
				set_lc_server();
				next_member=group->hash_table[hash];
			}
		}
		/* member is not available. we will permanently put that hash onto another server */
		if(status == -1) {
			/* then choose a new next_member */
			next_member++;
			next_member= next_member % group->nmembers;
			status=set_lc_server();
			if(status < 0) {
				write_log(OCTOPUS_LOG_STD, "WARNING: algorithm HASH: set_lc_server could not select any server");
				return status;
			}
			/* for accounting, decrement the dead server's amount of hashes */
			group->members[group->hash_table[hash]].hash_table_usage --;
			/* for accounting, the new server gets assigned another hash */
			group->members[next_member].hash_table_usage ++;
			/* finally, update the actual hash table */
			group->hash_table[hash]=next_member;

		}
		return 0;
	}
	/* HASH MISS: we haven't seen this hash val before, assign the least connected server */
	else {
		if(debug_level > 1) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: algorithm HASH: hash miss for uri %s", uri);
			write_log(OCTOPUS_LOG_STD, log_string);
		}
		/* when we have to choose a server, use LC as the selection algorithm. */
		status=set_lc_server();
		if(status < 0) {
			return status;
		}
		/*for accounting, the new server gets assigned another hash */
		group->members[next_member].hash_table_usage ++;
		/*finally, update the actual hash table */
		group->hash_table[hash]=next_member;
	}
	return 0;
}

int choose_ll_server(SESSION *session) {
	if(group->nmembers == 0) {
		return -1;
	}
	int status;
	status=set_ll_server();
	if(status < 0) {
		return status;
	}
	status = verify_selection();
	if(status < 0) {
		return status;
	}
	if(status == 1) {
		status=connect_server(session, NO_USE_CLONE);
		return status;
	}
	status=connect_server(session, USE_CLONE);
	return status;

}

int set_ll_server() {
	unsigned short int i;
	for (i=0; i < group->nmembers; i++) {
		/*if the candidate has exceeded its maximum load, or is invalid, then don't consider it */
		if((group->members[i].c >= group->members[i].maxc) || (group->members[i].status != SERVER_STATE_ENABLED)) {
			continue;
		}
		if((group->overload_mode == MODE_STRICT) && (group->members[i].e_load > 100)) {
			continue;
		}
		/*if the candidate has less load than the current next_member then it becomes the new next_member */
		/*e_load is not updated reguarly enough to be useful here */
		if((group->members[i].load + (group->members[i].c * group->session_weight)) < (group->members[next_member].load + group->members[next_member].c * group->session_weight)) {
			next_member=i;
		}
		/*if the current next_member is marked as disabled or failed then the candidate becomes the next_member */
		else if(group->members[next_member].status != SERVER_STATE_ENABLED) {
			next_member=i;
		}
		/*if the current next_member has exceeded its maximum load then the candidate becomes next_member */
		else if((group->members[next_member].e_load > 100) && (group->overload_mode == MODE_STRICT)){
			next_member=i;
		}
	}

	for(i=0;i<group->nclones;i++) {
		/*if the candidate has exceeded its maximum load, or is invalid, then don't consider it */
		if((group->clones[i].c >= group->clones[i].maxc) || (group->clones[i].status <= 0)) {
			continue;
		}
		if((group->overload_mode == MODE_STRICT) && (group->clones[i].e_load > 100)) {
			continue;
		}

		/*if the candidate has less load than the current next_clone then it becomes the new next_clone */
		/*e_load is not updated reguarly enough to be useful here */
		if((group->clones[i].load + (group->clones[i].c * group->session_weight)) < (group->clones[next_clone].load + group->clones[next_clone].c * group->session_weight)) {
			next_clone=i;
		}
		/*if the current next_clone is marked as disabled or failed then the candidate becomes the next_clone */
		else if(group->clones[next_clone].status != SERVER_STATE_ENABLED) {
			next_clone=i;
		}
		/*if the current next_clone has exceeded its maximum load then the candidate becomes next_clone */
		else if((group->clones[next_clone].e_load > 100) && (group->overload_mode == MODE_STRICT)) {
			next_clone=i;
		}
	}
	return 0;
}

/* sets the next_member and next_clone variables for the least connection algorithm */
int choose_lc_server(SESSION *session) {
	if(group->nmembers == 0) {
		return -1;
	}
	int status;
	status=set_lc_server();
	if(status < 0) {
		return status;
	}
	status = verify_selection();
	if(status < 0) {
		return status;
	}
	if(status == 1) {
		status=connect_server(session, NO_USE_CLONE);
		return status;
	}
	status=connect_server(session, USE_CLONE);
	return status;

}

/* sets the next_member and next_clone variables for the least connection algorithm */
/* the member and clone with the lowest number of active connections is the next_member and next_ghost respectively */
int set_lc_server() {
	unsigned short int i;
	if (group->nmembers > 0) {
		next_member++;
		next_member= next_member % group->nmembers;
	}
	else {
		return -1;
	}
	if (group->nclones > 0) {
		next_clone++;
		next_clone= next_clone % group->nclones;
	}
	for (i=0; i < group->nmembers; i++) {
		/*if the candidate has exceeded its max number of connections, or is invalid, then don't consider it */
		if((group->members[i].c >= group->members[i].maxc) || (group->members[i].status <= 0)) {
			continue;
		}
		/*if we are running in overload mode MODE_STRICT and the server is overloaded then skip it */
		if((group->overload_mode == MODE_STRICT) && (group->members[i].e_load > 100)) {
			continue;
		}
		/*if the candidate has less active connections than the current next_member and it is alive, then it becomes the new next_member */
		if(group->members[i].c < group->members[next_member].c) {
			next_member=i;
		}
		/*if the current next_member is marked as disabled or failed then the candidate becomes the next_member */
		else if(group->members[next_member].status != SERVER_STATE_ENABLED) {
			next_member=i;
		}
		/*if the current next_member has exceeded its maxc then the candidate becomes next_member */
		else if((group->members[next_member].c >= group->members[next_member].maxc)) {
			next_member=i;
		}
	}

	for(i=0;i<group->nclones;i++) {
		/*if the candidate has exceeded its max number of connections, or is invalid, then don't consider it */
		if((group->clones[i].c >= group->clones[i].maxc) || (group->clones[i].status <= 0)) {
			continue;
		}
		/*if we are running in overload mode MODE_STRICT and the server is overloaded then skip it */
		if((group->overload_mode == MODE_STRICT) && (group->clones[i].e_load > 100)) {
			continue;
		}
		/*if the candidate has less active connections than the current next_clone and it is alive, then it becomes the new next_clone */
		if(group->clones[i].c < group->clones[next_clone].c) {
			next_clone=i;
		}
		/*if the current next_clone is marked as disabled or failed then the candidate becomes the next_clone */
		else if(group->clones[next_clone].status != SERVER_STATE_ENABLED) {
			next_clone=i;
		}
		/*if the current next_clone has exceeded its maxc then the candidate becomes next_clone */
		else if((group->clones[next_clone].c >= group->clones[next_clone].maxc)) {
			next_clone=i;
		}
	}
	return 0;
}

/* sets the next_member and next_clone variables for the round robin algorithm */
int choose_rr_server(SESSION *session) {
	if(group->nmembers == 0) {
		return -1;
	}
	int status;
	status=set_rr_server();
	if(status < 0) {
		return status;
	}
	status = verify_selection();
	if(status < 0) {
		return status;
	}
	if(status == 1) {
		status=connect_server(session, NO_USE_CLONE);
		return status;
	}
	status=connect_server(session, USE_CLONE);
	return status;

}

/* sets the next_member and next_clone variables for the round robin algorithm */
/* returns 0 if both member and clone are valid */
/* returns 1 if the member is valid but the clone is not */
/* returns -1 if member is invalid */
int set_rr_server() {
	int i;
	for(i=0;i<group->nmembers;i++) {
		next_member++;
		next_member = next_member % (group->nmembers);
		/*if the candidate has exceeded its maximum load, or is invalid, then don't consider it */
		if((group->members[next_member].c >= group->members[next_member].maxc) || (group->members[next_member].status != SERVER_STATE_ENABLED)) {
			continue;
		}
		/*if we are running in overload mode MODE_STRICT and the server is overloaded then skip it */
		if((group->overload_mode == MODE_STRICT) && (group->members[next_member].e_load > 100)) {
			continue;
		}
		break;
	}
	for(i=0;i<group->nclones;i++) {
		next_clone ++;
		next_clone = next_clone % (group->nclones);
		/*if the candidate has exceeded its maximum load, or is invalid, then don't consider it */
		if((group->clones[next_clone].c >= group->clones[next_clone].maxc) || (group->clones[next_clone].status != SERVER_STATE_ENABLED)) {
			continue;
		}
		/*if we are running in overload mode MODE_STRICT and the server is overloaded then skip it */
		if((group->overload_mode == MODE_STRICT) && (group->clones[next_clone].e_load > 100)) {
			continue;
		}
		break;
	}
	return 0;
}


