/*
 * Octopus Load Balancer - Client binary.
 *
 * Copyright 2008 Alistair Reay <alreay1@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 */


#include "octopus.h"
int readonly;
int interactive;
int csv;

int usage(char *prog_name) {
	sockbufsize=0; /* just to suppress compiler 'unused variable' message */
	fprintf(stderr, "Octopus Client version " OCTOPUS_VERSION " date " OCTOPUS_DATE"\n");
	fprintf(stderr, "Copyright 2009 Alistair Reay <alreay1@gmail.com>\n\n");
	fprintf(stderr, "Usage : %s  [-f SHM_FILE] [-d SHM_FILE_DIR] [-e CMD]\n", prog_name);
	fprintf(stderr, "	-h 		Print help message\n");
	fprintf(stderr, "	-f <file>	Location of SHM file for a running instance\n");
	fprintf(stderr, "	-d <directory>	Directory to look for SHM files. Defaults to %s\n",DEFAULT_SHM_RUN_DIR);
	fprintf(stderr, "	-e <cmd>	Run a command then exit. Non-interactive mode\n");
	fprintf(stderr, "	-i		Ignore client/server version check. Dangerous option, allows a older/newer client to connect to server\n");
	fprintf(stderr, "	-r		Attach to server in read-only mode. This will be attempted automatically if r/w mode can't be set\n");
	fprintf(stderr, "	-m		Return output in csv format. Intended for interfaces to Octopus like web admin interface. Should only be used with option -e\n");
	fprintf(stderr, "\n");
	exit(0);
}

/* this function looks for .shm files in the supplied (or default) directory and then tries to connect client to instance */
int scan_for_balancers(char *run_dir, int ignore_version_check){
	struct stat stat_buffer;
	int status;
	int i;
	DIR *dp;
	struct dirent *ep;
	char *file_full_path;
	int counter=0;
	errno=0;
	char *balancers[MAXBALANCERS];
	char input[64];

	dp=opendir(run_dir);
	if(dp != NULL) {
		while((ep = readdir(dp))) {
			i=(strlen(ep->d_name) + strlen(run_dir) + 1);
			file_full_path=malloc(i * sizeof(char));
			snprintf(file_full_path, i, "%s%s", run_dir, ep->d_name);
			status = stat(file_full_path, &stat_buffer);
			if(S_ISREG(stat_buffer.st_mode)) {
				balancers[counter]= malloc(strlen(file_full_path));
				strncpy(balancers[counter], file_full_path, strlen(file_full_path));
				counter++;
			}
			free(file_full_path);
		}
		closedir(dp);
	}
	else {
		printf("ERROR: Could not read dir %s : %s\n", run_dir, strerror(errno));
	}
	if(counter > 1) {
		printf("select a balancer...\n");
		for(i=0;i<counter; i++) {
			printf("%d. %s\n",i+1,balancers[i]);
		}
		printf("\n> ");
		while(fgets(input, 8, stdin) == NULL) {
			continue;
		}
		if((atoi(input) > counter) || (atoi(input) <= 0)) {
			printf("ERROR: Invalid balancer %s\n", input);
			exit(1);
		}
		connect_to_shm(balancers[atoi(input)-1], ignore_version_check);

	}

	else if(counter == 1) {
		connect_to_shm(balancers[0], ignore_version_check);
	}
	else {
		printf("ERROR: Could not find any balancers\n");
		exit(1);
	}

	return 0;
}

/* this function does the work of actually connecting to the SHM segment of the server instance */
int connect_to_shm(char *run_file, int ignore_version_check) {

	char *data = NULL;
	key_t key;
	int shmid;
	int status;
	status = stat(run_file, &stat_buffer);

	if(status == -1) {
		fprintf(stderr, "ERROR: Unable to shm_run file. Check that server process has been started.\n");
		exit(-1);
	}
	if ((key = ftok(run_file, 'Z')) == -1) {
		fprintf(stderr, "ERROR: Unable to make SHM key\n");
		exit(-1);
	}
	if ((shmid = shmget(key, sizeof(GROUP), 006)) == -1) {
		if(errno == EACCES) {
			if ((shmid = shmget(key, sizeof(GROUP), 0)) == -1) {
				fprintf(stderr, "ERROR: Unable to connect to server process. Access error: %s\n", strerror(errno));
				exit(-1);
			}
			readonly=1;

		}
		else {
			fprintf(stderr, "ERROR: Unable to connect to server process. Check that server process has been started: %s\n", strerror(errno));
			exit(-1);
		}
	}
	if(readonly == 1) {
		data = shmat(shmid, (void *) 0, SHM_RDONLY);
	}
	else {
		data = shmat(shmid, (void *) 0, 0);
	}
	if (data == (char *) (-1)) {
		fprintf(stderr, "ERROR: Unable to attach to SHM segment: %s\n", strerror(errno));
		exit(-1);
	}
	group = (GROUP *)data;
	if(!(strncmp(group->version, OCTOPUS_VERSION, OCTOPUS_VERSION_LEN))) {
		if(csv==1) {
			printf("%s,%s:%d\n", group->version, inet_ntoa(group->binding_ip),group->binding_port);
		}
		else {
			printf("Connected to balancer at %s:%d\n", inet_ntoa(group->binding_ip),group->binding_port);
		}
	}
	else {
		if(ignore_version_check > 0) {
			printf("WARNING! Version mismatch: server is version %s, client is version %s. Continuing as client was started with -i option\n", group->version, OCTOPUS_VERSION);
		}
		else {
			printf("ERROR! Version mismatch: server is version %s, client is version %s\n", group->version, OCTOPUS_VERSION);
			exit(-1);
		}
	}
	if(group->alive==0) {
		printf("ERROR! Connect succeeded but SHM segment has been marked as dead. Quitting\n");
		exit(-1);
	}
	return 0;
}


int main(int argc, char *argv[]) {
	int i;
	int status;
	char input[CLIENT_MAX_INPUT];
	char *cli_input=NULL;
	char *c1;
	char *c2;
	char *c3;
	char *c4;
	char *c5;
	int v1;
	int v2;
	float f1;
	int ignore_version_check=0;
	char *run_file = "";
	char *run_dir = DEFAULT_SHM_RUN_DIR;
	int use_run_file=0;
	readonly=0;
	interactive=1;
	csv=0;

	while((i = getopt(argc, argv,"mirhd:f:e:")) != -1) {
		switch (i) {
			case 'd':
				run_dir = optarg;
				break;
			case 'f':
				run_file=optarg;
				use_run_file=1;
				break;
			case 'e':
				cli_input=optarg;
				interactive=-1;
				break;
			case 'h':
				usage(argv[0]);
				exit(0);
				break;
			case 'r':
				readonly=1;
				break;
			case 'i':
				ignore_version_check=1;
				break;
			case 'm':
				csv=1;
				break;
			case '?':
				if(isprint(optopt)) {
					fprintf(stderr, "ERROR: main: Unknown option `-%c'.\n",optopt);
				}
				else {
					fprintf(stderr, "ERROR: main: Unknown option character `\\x%x'.\n",optopt);
				}
				return 1;
			default:
				abort();
		}
	}
	for(i = optind; i < argc; i++) {
		fprintf(stderr, "ERROR: main: Non-option argument %s\n", argv[i]);
		exit(1);
	}
	/*if user has supplied a specific .shm run file location then we'll just use that and not look for any others */
	if(use_run_file > 0) {
		connect_to_shm(run_file, ignore_version_check);
	}
	else {
		scan_for_balancers(run_dir, ignore_version_check);
	}

	if(interactive==1) {
		if(readonly == 1) {
			printf("Octopus Interactive Shell (read-only mode)\n");
		}
		else {
			printf("Octopus Interactive Shell\n");
		}
		printf("enter '?' for help\n");
	}
	while(1) {
		if(interactive==0) {
			exit(1);
		}
		if(interactive==1) {
			if(readonly == 1) {
				printf("\noctopus (read-only mode)$ ");
			}
			else {
				printf("\noctopus $ ");
			}
		}
		if(interactive == -1) {
			strncpy(input, cli_input, CLIENT_MAX_INPUT);
			if(strlen(input) >= CLIENT_MAX_INPUT - 1) {
				printf("ERROR: Too much input!\n");
				exit(1);
			}
			/* set this counter to zero so that after processing the current command the client interface exits */
			interactive=0;
		}
		else {
			if(fgets(input, CLIENT_MAX_INPUT, stdin) == NULL) {
				continue;
			}
		}
		if(group->alive != 1) {
			printf("NOTICE: Server process has quit. Exitting.\n");
			exit(0);
		}
		if(strlen(input) >= (CLIENT_MAX_INPUT -1) ) {
			printf("ERROR: Too much input!\n");
			exit(1);
		}
		/*break the user's input into separate pieces */
		c1 = strtok(input," ");
		c2 = strtok(NULL, " ");
		c3 = strtok(NULL, " ");
		c4 = strtok(NULL, " ");
		c5 = strtok(NULL, " ");
		if(!strcmp(c1, "\n")) {
			continue;
		}

		if(!strncmp(c1, "scan", 4)) {
			readonly=0;
			scan_for_balancers(run_dir, ignore_version_check);
		}

		else if(!strncmp(c1, "snmp", 4)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			if(group->snmp_status==SNMP_NOT_INCLUDED) {
				printf("ERROR: server instance does not have SNMP support compiled!\n");
			}
			else {
				if(c2 == NULL) {
					printf("ERROR: not enough args\n");
					continue;
				}
				if(!strncmp(c2, "on",2 )) {
					group->snmp_status=SNMP_ENABLED;
					printf("Set snmp status to ENABLED");
				}
				if(!strncmp(c2, "off",3 )) {
					group->snmp_status=SNMP_DISABLED;
					printf("Set snmp status to DISABLED");
				}
			}
			continue;
		}

		else if(!strncmp(c1, "s",1)) {
			if(csv==1) {
				for(i=0; i<group->nmembers; i++) {
					if(group->members[i].status == SERVER_STATE_FREE) {
						continue;
					}
					printf("%s,%d,%s,%s,%s,%d,%d,%d,%lu,%lu,%lu,%f,%f,%d\n","Member", i, group->members[i].name, server_status[group->members[i].status + 3], inet_ntoa(group->members[i].myaddr.sin_addr), group->members[i].port, group->members[i].c, group->members[i].maxc, group->members[i].completed_c, group->members[i].bsent, group->members[i].brecv, group->members[i].load, group->members[i].maxload, group->members[i].e_load);
				}
				for(i=0; i<group->nclones; i++) {
					if(group->clones[i].status == SERVER_STATE_FREE) {
						continue;
					}
					printf("%s,%d,%s,%s,%s,%d,%d,%d,%lu,%lu,%lu,%f,%f,%d\n","Clone", i, group->clones[i].name, server_status[group->clones[i].status + 3], inet_ntoa(group->clones[i].myaddr.sin_addr), group->clones[i].port, group->clones[i].c, group->clones[i].maxc, group->clones[i].completed_c, group->clones[i].bsent, group->clones[i].brecv, group->clones[i].load, group->clones[i].maxload, group->clones[i].e_load );
				}
			}
			else {
				if(group->overall_load >= 0) {
					printf("Balancer @ %s:%d, using algorithm %s with clone mode %s and overload mode %s. Cluster overall load @ %d%%\n", inet_ntoa(group->binding_ip), group->binding_port, algorithm_status[group->algorithm - 1], cloning_status[group->clone_mode + 2], overload_status[group->overload_mode - 1], group->overall_load);
				}
				else {
					printf("Balancer @ %s:%d, using algorithm %s with clone mode %s and overload mode %s\n", inet_ntoa(group->binding_ip), group->binding_port, algorithm_status[group->algorithm - 1], cloning_status[group->clone_mode + 2], overload_status[group->overload_mode - 1]);
				}
				if(group->snmp_status==SNMP_ENABLED) {
					printf("%7s %3s %16s  %8s %16s %4s %5s %5s %7s %12s %12s %5s %8s %7s\n","Type", "#", "Name", "Status", "ip-address", "port", "c", "maxc", "hc", "bsent", "brecv", "load", "max-load", "loading");
					for(i=0; i<group->nmembers; i++) {
						if(group->members[i].status == SERVER_STATE_FREE) {
							continue;
						}
						if(group->members[i].load < 0) {
							printf("%7s %3d %16s  %8s %16s %4d %5d %5d %7lu %12lu %12lu  %4s  %2.2f\n","Member", i, group->members[i].name, server_status[group->members[i].status + 3], inet_ntoa(group->members[i].myaddr.sin_addr), group->members[i].port, group->members[i].c, group->members[i].maxc, group->members[i].completed_c, group->members[i].bsent, group->members[i].brecv, "ERR", group->members[i].maxload);
						}
						else {
							printf("%7s %3d %16s  %8s %16s %4d %5d %5d %7lu %12lu %12lu  %2.2f  %2.2f     %3d%%\n","Member", i, group->members[i].name, server_status[group->members[i].status + 3], inet_ntoa(group->members[i].myaddr.sin_addr), group->members[i].port, group->members[i].c, group->members[i].maxc, group->members[i].completed_c, group->members[i].bsent, group->members[i].brecv, group->members[i].load, group->members[i].maxload, group->members[i].e_load);
						}
					}
					for(i=0; i<group->nclones; i++) {
						if(group->clones[i].status == SERVER_STATE_FREE) {
							continue;
						}
						if(group->clones[i].load < 0) {
							printf("%7s %3d %16s  %8s %16s %4d %5d %5d %7lu %12lu %12lu  %4s  %2.2f\n","Clone", i, group->clones[i].name, server_status[group->clones[i].status + 3], inet_ntoa(group->clones[i].myaddr.sin_addr), group->clones[i].port, group->clones[i].c, group->clones[i].maxc, group->clones[i].completed_c, group->clones[i].bsent, group->clones[i].brecv, "ERR", group->clones[i].maxload);
						}
						else {
							printf("%7s %3d %16s  %8s %16s %4d %5d %5d %7lu %12lu %12lu  %2.2f  %2.2f     %3d%%\n","Clone", i, group->clones[i].name, server_status[group->clones[i].status + 3], inet_ntoa(group->clones[i].myaddr.sin_addr), group->clones[i].port, group->clones[i].c, group->clones[i].maxc, group->clones[i].completed_c, group->clones[i].bsent, group->clones[i].brecv, group->clones[i].load, group->clones[i].maxload, group->clones[i].e_load );
						}
					}
				}
				else {
					printf("%7s %3s %16s  %8s %16s %4s %5s %5s %7s %12s %12s\n","Type", "#", "Name", "Status", "ip-address", "port", "c", "maxc", "hc", "bsent", "brecv");
					for(i=0; i<group->nmembers; i++) {
						if(group->members[i].status == SERVER_STATE_FREE) {
							continue;
						}
						printf("%7s %3d %16s  %8s %16s %4d %5d %5d %7lu %12lu %12lu\n","Member", i, group->members[i].name, server_status[group->members[i].status + 3], inet_ntoa(group->members[i].myaddr.sin_addr), group->members[i].port, group->members[i].c, group->members[i].maxc, group->members[i].completed_c, group->members[i].bsent, group->members[i].brecv);
					}
					for(i=0; i<group->nclones; i++) {
						if(group->clones[i].status == SERVER_STATE_FREE) {
							continue;
						}
						printf("%7s %3d %16s  %8s %16s %4d %5d %5d %7lu %12lu %12lu\n","Clone", i, group->clones[i].name, server_status[group->clones[i].status + 3], inet_ntoa(group->clones[i].myaddr.sin_addr), group->clones[i].port, group->clones[i].c, group->clones[i].maxc, group->clones[i].completed_c, group->clones[i].bsent, group->clones[i].brecv);
					}
				}
			}
		}
		else if(!strncmp(c1, "?",1)) {
			if(readonly==1) {
				printf("available commands for read-only mode:\n");
				printf("[s]how						overview and statistics for all members and clones\n");
				printf("[i]nfo						overview of load balancer configuration\n");
				printf("[scan]						search for and connect to other instances of octopus running on same host\n");
				printf("[q]uit						quit the client interface\n");
				printf("[ex]amples					show some usage examples\n");
				printf("[?]						this screen\n");
			}
			else {
				printf("available commands:\n");
				printf("[s]how						overview and statistics for all member and clone servers\n");
				printf("[c]reate <[c]lone/[m]ember> <name> <ip> <port>	create a new member or clone server\n");
				printf("[delete] <[c]lone/[m]ember> <#>			deletes specified member or clone server\n");
				printf("[d]isable [a]ll / (<[c]lone/[m]ember> <#>)	disables all (or specified members or clone) servers\n");
				printf("[e]nable [a]ll / (<[c]lone/[m]ember> <#>)	enables all (or specified member or clone) servers\n");
				printf("[maxc] <[c]lone/[m]ember> <#> <value>		sets the maximum number of connections for the specified member or clone\n");
				if(group->snmp_status != SNMP_NOT_INCLUDED) {
					printf("[maxl] <[c]lone/[m]ember> <#> <value>		sets the maximum load for the specified member or clone\n");
					printf("[o]verload <[s]trict/[r]elaxed>			set the overload mode\n");
					printf("[a]lgorithm <[rr]/[lc]/[ll]/[hash]>		set the balancing algorithm\n");
					printf("[hrt] <value>	  				set the hash rebalance threshold to <value>%%\n");
					printf("[hrs] <value>  					set the hash rebalance size to <value>%%\n");
					printf("[hri] <value>	  				set the hash rebalance interval to <value> seconds\n");
					printf("[snmp] <on/off>					set snmp monitoring on or off\n");
				}
				else {
					printf("[a]lgorithm <[rr]/[lc]/[hash]>			set the balancing algorithm\n");
				}
				printf("[m]ode <[s]trict/[f]ast/[o]ff>			set the cloning mode\n");
				printf("[monitor] <seconds>				set the time period between runs of the monitor process\n");
				printf("[r]eset <[a]ll> / <[c]lone/[m]ember> <#>	resets counters for member, clone or all servers\n");
				printf("[i]nfo						overview of load balancer configuration\n");
				printf("[scan]						search for and connect to other instances of octopus running on same host\n");
				printf("[q]uit						quit the client interface\n");
				printf("[kill]						kill server process (and client)\n");
				printf("[ex]amples					show some usage examples\n");
				printf("[?]						this screen\n");
			}

		}
		else if(!strncmp(c1, "exit",4)) {
			printf("Bye!\n");
			exit(0);
		}
		else if(!strncmp(c1, "q",1)) {
			printf("Bye!\n");
			exit(0);
		}
		else if(!strncmp(c1, "ex",2)) {
			printf("Usage Examples:\n\n");
			printf("Showing the current configuration including server names, IP, port, server type, and id #\n");
			printf("$ show\n");
			printf("\n");
			printf("Creating a new member webserver named \"Hercules\" @ 192.168.1.102 on tcp/80:\n");
			printf("$ create member Hercules 192.168.1.102 80\n");
			printf("\n");
			printf("Setting the maximum number of connections we will allow to \"Hercules\" which has id # of 2\n");
			printf("$ maxc m 2 500\n");
			printf("\n");
			printf("Setting the maximum load we will allow for \"Hercules\" which has id # of 2\n");
			printf("$ maxl m 2 4.0\n");
			printf("\n");
			printf("Creating a new clone webserver named \"David\" @ 10.1.1.5 on tcp/80:\n");
			printf("$ create clone David 10.1.1.5 80\n");
			printf("\n");
			printf("Turning on request cloning in strict mode\n");
			printf("$ mode strict\n");
			printf("\n");
			printf("Setting overload to strict mode\n");
			printf("$ overload strict\n");
			printf("\n");
			printf("Dynamically changing the load balancing algorithm to \"Least Connections\"\n");
			printf("$ algorithm lc\n");
			printf("\n");


		}


		else if(!strncmp(c1, "delete", 6)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			v1=strtol(c2, &c1, 10);
			if(c2 != c1) {
				/*user has not specified whether number is for members or clones. we assume member */
				if(v1 < group->nmembers) {
					printf("Deleting member %s\n", group->members[v1].name);
					group->members[v1].status = SERVER_STATE_DELETED;
					continue;
				}
				else {
					printf("ERROR: member id not in range!\n");
					continue;
				}
			}
			else {
				if(c3 == NULL) {
					printf("ERROR: not enough args\n");
					continue;
				}
				v1=strtol(c3,&c1, 10);
				if(c3 == c1) {
					printf("ERROR: argument 3 is not a number\n");
					continue;
				}
				if(!strncmp(c2, "m",1)) {
					/*user wants to delete a member server */
					if(v1 <= group->nmembers) {
						printf("Deleting member %s\n", group->members[v1].name);
						group->members[v1].status = SERVER_STATE_DELETED;
						continue;
					}
					else {
						printf("ERROR: member id not in range!\n");
						continue;
					}
				}
				else if(!strncmp(c2, "c",1)) {
					/*user wants to delete a clone server */
					if(v1 <= group->nclones) {
						printf("Deleting clone %s\n", group->clones[v1].name);
						group->clones[v1].status = SERVER_STATE_DELETED;
						continue;
					}
					else {
						printf("ERROR: clone id not in range!\n");
						continue;
					}
				}
				else {
					printf("ERROR: argument 2 invalid!\n");
				}
			}

		}
		else if(!strncmp(c1, "monitor",7)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			v1=strtol(c2, &c1, 10);
            if(c2 != c1) {
				if(v1 < 0) {
	            	printf("ERROR: argument 2 invalid\n");
					continue;
				}
				else {
					group->monitor_interval=v1;
					printf("Set monitor interval to %d seconds\n", group->monitor_interval);
				}
			}
            else {
            	printf("ERROR: argument 2 invalid\n");
				continue;
            }

		}
		else if(!strncmp(c1, "hrt",3)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			v1=strtol(c2, &c1, 10);
            if(c2 != c1) {
				if((v1 < 0) || (v1 > 100)) {
	            	printf("ERROR: argument 2 invalid\n");
					continue;
				}
				else {
					group->hash_rebalance_threshold=v1;
					printf("Set hash rebalance threshold to %d\n", group->hash_rebalance_threshold);
				}
			}
            else {
            	printf("ERROR: argument 2 invalid\n");
				continue;
            }
		}

		else if(!strncmp(c1, "hrs",3)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			v1=strtol(c2, &c1, 10);
            if(c2 != c1) {
				if((v1 < 0) || (v1 > 100)) {
	            	printf("ERROR: argument 2 invalid\n");
					continue;
				}
				else {
					group->hash_rebalance_size=v1;
					printf("Set hash rebalance size to %d%%\n", group->hash_rebalance_size);
				}
			}
            else {
            	printf("ERROR: argument 2 invalid\n");
				continue;
            }
		}

		else if(!strncmp(c1, "hri",3)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			v1=strtol(c2, &c1, 10);
            if(c2 != c1) {
				if(v1 <= 0) {
	            	printf("ERROR: argument 2 invalid\n");
					continue;
				}
				else {
					group->hash_rebalance_interval=v1;
					if(v1==0) {
						printf("Disabled hash rebalancing\n");
					}
					else {
						printf("Set hash rebalance interval to %d seconds\n", group->hash_rebalance_interval);
					}
				}
			}
            else {
            	printf("ERROR: argument 2 invalid\n");
				continue;
            }
		}
		else if(!strncmp(c1, "i",1)) {
			printf("Octopus version : %s\n", group->version);
			printf("Process info\n");
			printf("Balancer PID:		%d\n", group->master_pid);
			printf("Monitor PID:		%d\n", group->monitor_pid);
			printf("Shared Memory ID:	%d\n", group->shmid);
			printf("Shared Memory file:	%s\n", group->shm_run_file_fullname);
			printf("\n");

			printf("Network info\n");
			printf("Listening IP:		%s\n", inet_ntoa(group->binding_ip));
			printf("Listening TCP port:	%d\n", group->binding_port);
			printf("Member outbound IP:	%s\n", inet_ntoa(group->member_outbound_ip));
			printf("Clone outbound IP:	%s\n", inet_ntoa(group->clone_outbound_ip));
			printf("File Descriptor Limit:	%d\n", group->fd_limit);
			printf("Session Limit:		%d\n", group->session_limit);
			printf("\n");

			printf("Balancing algorithm info\n");
			printf("Algorithm:		%s\n", algorithm_status[group->algorithm - 1]);
			printf("Clone mode:		%s\n", cloning_status[group->clone_mode + 2]);
			printf("Overload mode:		%s\n", overload_status[group->overload_mode - 1]);
			printf("Session Weight:		%f\n", group->session_weight);
			printf("Default member/clone maximum connection limit:	%d\n", group->default_maxc);
			printf("Default member/clone maximum load limit:	%f\n", group->default_maxl);
			printf("\n");

			printf("Monitor info:\n");
			printf("Interval:			%d\n", group->monitor_interval);
			printf("Check timeout:			%d\n", group->connect_timeout);
			if(group->snmp_status == SNMP_ENABLED) {
				printf("SNMP status:			Enabled\n");
			}
			if(group->snmp_status == SNMP_DISABLED) {
				printf("SNMP status:			Disabled\n");
			}
			if(group->snmp_status == SNMP_NOT_INCLUDED) {
				printf("SNMP status:			Not compiled!\n");
			}
			printf("\n");

			printf("URI Hash info:\n");
			printf("Table size:			%d\n", HASHTABLESIZE);
			printf("Rebalance threshold %%:	 	%d\n", group->hash_rebalance_threshold);
			printf("Rebalance size %%: 		%d\n", group->hash_rebalance_size);
			printf("Rebalance interval (secs): 	%d\n", group->hash_rebalance_interval);
			printf("%16s %10s\n", "Server Name", "URI Assignments");
			for(i=0; i<group->nmembers; i++) {
				if(group->members[i].status != SERVER_STATE_FREE) {
					printf("%16s %8d\n", group->members[i].name, group->members[i].hash_table_usage);
				}
			}
		}
		/* ALGORITHM change */
		else if(!strncmp(c1, "a",1)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			else if(!strncmp(c2, "rr",2)) {
				group->algorithm=ALGORITHM_RR;
				printf("Set balancing algorithm to roundrobin\n");
				continue;
			}
			else if(!strncmp(c2, "lc",2)) {
				group->algorithm=ALGORITHM_LC;
				printf("Set balancing algorithm to least connections\n");
				continue;
			}
			else if(!strncmp(c2, "ll",2)) {
				#ifdef USE_SNMP
				group->algorithm=ALGORITHM_LL;
				printf("Set balancing algorithm to least load\n");
				#else
				printf("ERROR: Cannot use Lease Load algorithm unless SNMP support is enabled at compile time\n");
				#endif
				continue;
			}
			else if(!strncmp(c2, "hash",4)) {
				group->algorithm=ALGORITHM_HASH;
				printf("Set balancing algorithm to uri hashing\n");
				continue;
			}
			else {
				printf("ERROR: Argument 2 invalid\n");
				continue;
			}

		}
		/* KILL command */
		else if(!strncmp(c1, "kill", 4)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			printf("Killing server with master pid %d and monitor pid %d\n", group->master_pid, group->monitor_pid);
			kill(group->monitor_pid, SIGTERM);
			kill(group->master_pid, SIGTERM);
			exit(0);
		}
		else if(!strncmp(c1, "c",1)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			errno=0;
			int cloneServer=0;
			int serverStatus=0;
			int serverPort=0;
			int serverMaxc=group->default_maxc;
			float serverMaxload=group->default_maxl;
			char serverName[SERVERNAME_MAX_LENGTH];
			struct in_addr serverIP;
			memset(serverName,'\0',SERVERNAME_MAX_LENGTH);
			/*validate the inputs */
			if((c2 == NULL) || (c3 == NULL) || (c4 == NULL) || (c5 == NULL)) {
				printf("ERROR: not enough args\n");
				continue;
			}
			if(strlen(c4) > 15) {
				printf("ERROR: ip value too long\n");
				continue;
			}
			/*member, clone or invalid? */
			if(!strncmp(c2,"m",1)) {
				cloneServer=0;
			}
			else if(!strncmp(c2,"c",1)) {
				cloneServer=1;
			}
			else {
				printf("ERROR: argument 2 invalid\n");
				continue;
			}
			strncpy(serverName, c3, SERVERNAME_MAX_LENGTH);
			if(inet_aton(c4 ,&serverIP) ==0) {
				printf("ERROR: invalid ip %s for server %s\n",c4,serverName);
				continue;
			}
			serverPort=atoi(c5);
			if (!(serverPort > 0)) {
				printf("ERROR: port value invalid\n");
				continue;
			}
			/*ok we've validated our input now make a new server */
			status=create_group_server(cloneServer, serverName, serverStatus, serverPort, &serverIP, serverMaxc, serverMaxload);
			if(status==0) {
				printf("created server \"%s\" successfully\n",serverName);
				continue;
			}
			else {
				printf("ERROR: unable to create server \"%s\"\n", serverName);
				continue;
			}


		}
		/* MAXL command */
		else if(!strncmp(c1, "maxl",4)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			errno=0;
			f1=0.0;
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			v1=strtol(c2, &c1, 10);
			if(c2 != c1) {
				/*user has not specified whether number is for members or clones. we assume member */
				f1=atof(c3);
				if(v1 < group->nmembers) {
					group->members[v1].maxload=f1;
					printf("Set member %d max load to %2.2f\n",v1, f1);
					continue;
				}
				else {
					printf("ERROR: member id not in range!\n");
					continue;
				}
			}
			else {
				if(c3 == NULL) {
					printf("ERROR: not enough args\n");
					continue;
				}
				v1=strtol(c3,&c1, 10);
				if(c3 == c1) {
					printf("ERROR: argument 3 is not a number\n");
					continue;
				}
				f1=atof(c4);
				if(!strncmp(c2, "m",1)) {
					/*user wants to set maxc for a member server */
					if(v1 <= group->nmembers) {
						group->members[v1].maxload=f1;
						printf("Set member %d max load to %f\n",v1, f1);
						continue;
					}
					else {
						printf("ERROR: member id not in range!\n");
						continue;
					}
				}
				else if(!strncmp(c2, "c",1)) {
					/*user wants to set maxc for a clone server */
					if(v1 <= group->nclones) {
						group->clones[v1].maxload=f1;
						printf("Set clone %d max connections to %f\n",v1, f1);
						continue;
					}
					else {
						printf("ERROR: clone id not in range!\n");
						continue;
					}
				}
				else {
					printf("ERROR: argument 2 invalid!\n");
				}
			}
		}
		/* MAXC command */
		else if(!strncmp(c1, "maxc",4)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			errno=0;
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			v1=strtol(c2, &c1, 10);
			if(c2 != c1) {
				/*user has not specified whether number is for members or clones. we assume member */
				v2=strtol(c3, &c1, 10);
				if(c3 != c1) {
					if(v1 < group->nmembers) {
						group->members[v1].maxc=v2;
						printf("Set member %d max connections to %d\n",v1, v2);
						continue;
					}
					else {
						printf("ERROR: member id not in range!\n");
						continue;
					}
				}
				else {
					printf("ERROR: argument 3 invalid\n");
				}


			}
			else {
				if(c3 == NULL) {
					printf("ERROR: not enough args\n");
					continue;
				}
				v1=strtol(c3,&c1, 10);
				if(c3 == c1) {
					printf("ERROR: argument 3 is not a number\n");
					continue;
				}
				v2=strtol(c4,&c1, 10);
				if(c4 == c1) {
					printf("ERROR: argument 4 is not a number\n");
					continue;
				}
				if(!strncmp(c2, "m",1)) {
					/*user wants to set maxc for a member server */
					if(v1 <= group->nmembers) {
						group->members[v1].maxc=v2;
						printf("Set member %d max connections to %d\n",v1, v2);
						continue;
					}
					else {
						printf("ERROR: member id not in range!\n");
						continue;
					}
				}
				else if(!strncmp(c2, "c",1)) {
					/*user wants to set maxc for a clone server */
					if(v1 <= group->nclones) {
						group->clones[v1].maxc=v2;
						printf("Set clone %d max connections to %d\n",v1, v2);
						continue;
					}
					else {
						printf("ERROR: clone id not in range!\n");
						continue;
					}
				}
				else {
					printf("ERROR: argument 2 invalid!\n");
				}
			}

		}

		/* RESET command */
		else if(!strncmp(c1, "r",1)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			if(!strncmp(c2, "all", 1)) {
				for(i=0;i< group->nmembers; i++) {
					group->members[i].bsent=0;
					group->members[i].brecv=0;
					group->members[i].completed_c=0;
				}
				for(i=0;i< group->nclones; i++) {
					group->clones[i].bsent=0;
					group->clones[i].brecv=0;
					group->clones[i].completed_c=0;
				}
				printf("Reset all counters\n");
				continue;
			}
			v1=strtol(c2, &c1, 10);
			if(c2 != c1) {
				/*user has not specified whether number is for members or clones. we assume member */
				if(v1 < group->nmembers) {
					group->members[v1].bsent=0;
					group->members[v1].brecv=0;
					group->members[v1].completed_c=0;
					continue;
				}
				else {
					printf("ERROR: member id not in range!\n");
					continue;
				}


			}
			else {
				if(c3 == NULL) {
					printf("ERROR: not enough args\n");
					continue;
				}
				v1=strtol(c3,&c1, 10);
				if(c3 == c1) {
					printf("ERROR: argument 3 is not a number\n");
					continue;
				}
				if(!strncmp(c2, "m",1)) {
					/*user wants to reset a member server */
					if(v1 <= group->nmembers) {
						group->members[v1].bsent=0;
						group->members[v1].brecv=0;
						group->members[v1].completed_c=0;
						printf("Reset counters for member %s\n", group->members[v1].name);
						continue;
					}
					else {
						printf("ERROR: member id not in range!\n");
						continue;
					}
				}
				else if(!strncmp(c2, "c",1)) {
					/*user wants to reset a clone server */
					if(v1 <= group->nclones) {
						group->clones[v1].bsent=0;
						group->clones[v1].brecv=0;
						group->clones[v1].completed_c=0;
						printf("Reset counters for clone %s\n", group->clones[v1].name);
						continue;
					}
					else {
						printf("ERROR: clone id not in range!\n");
						continue;
					}
				}
				else {
					printf("ERROR: argument 2 invalid!\n");
				}
			}

		}
		/* ENABLE command */
		else if(!strncmp(c1, "e",1)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			/* user has put in "enable all" */
			if(!strncmp(c2, "a",1)) {
				for(i=0;i<group->nmembers;i++) {
					if(group->members[i].status==SERVER_STATE_DISABLED) {
						group->members[i].status=SERVER_STATE_ENABLED;
						printf("Enabling member %s\n",group->members[i].name);
					}
				}
				for(i=0;i<group->nclones;i++) {
					if(group->clones[i].status==SERVER_STATE_DISABLED) {
						group->clones[i].status=SERVER_STATE_ENABLED;
						printf("Enabling clone %s\n",group->clones[i].name);
					}
				}
				continue;
			}
			v1=strtol(c2,&c1, 10);
			if(c2 != c1) {
				/*user has not specified whether number is for members or clones. we assume member */
				if(v1 < group->nmembers) {
					group->members[v1].status=SERVER_STATE_ENABLED;
					printf("Enabling member %s\n",group->members[v1].name);
				}
				else {
					printf("ERROR: member id not in range!\n");
					continue;
				}
			}
			else {
				if(c3 == NULL) {
					printf("ERROR: not enough args\n");
					continue;
				}
				v1=strtol(c3,&c1, 10);
				if(c3 == c1) {
					printf("ERROR: argument 3 is not a number\n");
					continue;
				}
				if(!strncmp(c2, "m",1)) {
					/*user wants to enable a member server */
					if(v1 < group->nmembers) {
						group->members[v1].status=SERVER_STATE_ENABLED;
						printf("Enabling member %s\n",group->members[v1].name);
					}
					else {
						printf("ERROR: member id not in range!\n");
						continue;
					}
				}
				else if(!strncmp(c2, "c",1)) {
					/*user wants to enable a clone server */
					if(v1 < group->nclones) {
						group->clones[v1].status=SERVER_STATE_ENABLED;
						/*enabling a server may mean setting clone mode from failed to enabled */
						set_cloned_state();
						printf("Enabling clone %s\n",group->clones[v1].name);
					}
					else {
						printf("ERROR: clone id not in range!\n");
						continue;
					}
				}
				else {
					printf("ERROR: argument 2 invalid!\n");
				}
			}
		}
		else if(!strncmp(c1, "m",1)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			if(!strncmp(c2, "s",1)) {
				group->clone_mode=MODE_STRICT;
				set_cloned_state();
				printf("Set clone mode to STRICT\n");
				continue;
			}
			else if(!strncmp(c2, "f",1)) {
				group->clone_mode=MODE_FAST;
				set_cloned_state();
				printf("Set clone mode to FAST\n");
				continue;
			}
			else if(!strncmp(c2, "o",1)) {
				group->clone_mode=MODE_OFF;
				set_cloned_state();
				printf("Set clone mode to OFF\n");
				continue;
			}
			else {
				printf("ERROR: Argument 2 invalid\n");
				continue;
			}
		}
		else if(!strncmp(c1, "o",1)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			if(!strncmp(c2, "r",1)) {
				group->overload_mode=MODE_RELAXED;
				printf("Set overload mode to RELAXED\n");
				continue;
			}
			else if(!strncmp(c2, "s",1)) {
				group->overload_mode=MODE_STRICT;
				printf("Set overload mode to STRICT\n");
				continue;
			}
			else {
				printf("ERROR: Argument 2 invalid\n");
				continue;
			}
		}
		else if(!strncmp(c1, "d",1)) {
			if(readonly==1) {
				printf("ERROR: This command not available in read-only mode!\n");
				continue;
			}
			if(c2 == NULL) {
				printf("ERROR: not enough args\n");
				continue;
			}
			/* user has put in "disable all" */
			if(!strncmp(c2, "a",1)) {
				for(i=0;i<group->nmembers;i++) {
					if((group->members[i].status==SERVER_STATE_ENABLED) || (group->members[i].status==SERVER_STATE_FAILED)) {
						group->members[i].status=SERVER_STATE_DISABLED;
					}
				}
				for(i=0;i<group->nclones;i++) {
					if((group->clones[i].status==SERVER_STATE_ENABLED) || (group->clones[i].status==SERVER_STATE_FAILED)) {
						group->clones[i].status=SERVER_STATE_DISABLED;
					}
				}
				continue;
			}
			v1=strtol(c2,&c1, 10);
			if(c2 != c1) {
				/*user has not specified whether number is for members or clones. we assume member */
				if(v1 < group->nmembers) {
					group->members[v1].status=SERVER_STATE_DISABLED;
					printf("Disabling member %s\n",group->members[v1].name);
				}
				else {
					printf("ERROR: member id not in range!\n");
					continue;
				}
			}
			else {
				if(c3 == NULL) {
					printf("ERROR: not enough args\n");
					continue;
				}
				v1=strtol(c3,&c1, 10);
				if(c3 == c1) {
					printf("ERROR: argument 3 is not a number\n");
					continue;
				}
				if(!strncmp(c2, "m",1)) {
					/*user wants to disable a member server */
					if(v1 < group->nmembers) {
						group->members[v1].status=SERVER_STATE_DISABLED;
						printf("Disabling member %s\n",group->members[v1].name);
					}
					else {
						printf("ERROR: member id not in range!\n");
						continue;
					}
				}
				else if(!strncmp(c2, "c",1)) {
					/*user wants to disable a member server */
					if(v1 < group->nclones) {
						group->clones[v1].status=SERVER_STATE_DISABLED;
						set_cloned_state();
						printf("Disabling clone %s\n",group->clones[v1].name);
					}
					else {
						printf("ERROR: clone id not in range!\n");
						continue;
					}
				}
				else {
					printf("ERROR: argument 2 invalid!\n");
				}
			}
		}
		else {
			printf("unrecognized command\n");
		}
	}
}
