/*
 * Octopus Load Balancer - Initialization functions.
 *
 * Copyright 2008 Alistair Reay <alreay1@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 */

/* runs the server process in the background (daemon mode) */
int initialize_process() {
	if(group->foreground == FOREGROUND_INIT) {
		pid_t pid;
		pid_t sid;
		if ((pid = fork()) < 0) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_process: Cannot fork a daemon - %s\n", strerror(errno));
			write_log(OCTOPUS_LOG_EXIT, log_string);
		}
		else if(pid != 0) {
			exit(0);
		}
		umask(0);
		sid = setsid();
		if (sid < 0) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_process: Cannot set sid - %s\n", strerror(errno));
			write_log(OCTOPUS_LOG_EXIT, log_string);
		}
		if((chdir("/")) < 0) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_process: Cannot chdir to / - %s\n", strerror(errno));
			write_log(OCTOPUS_LOG_EXIT, log_string);
		}
		fclose(stdin);
		fclose(stdout);
	}

	group->master_pid= getpid();
	return 0;
}

/* Parse the configuration file looking for the logfile directive */
int initialize_logging(char *conf_file) {
	char *new_log_file=NULL;
	FILE *fp;
	char *directive;
	char *value;
	char buffer[MAX_CONF_LINE_LEN];
	int lineCounter=0;
	FILE *log;
	errno=0;

	if((fp=fopen(conf_file,"r"))!=NULL) {
		while(fgets(buffer, MAX_CONF_LINE_LEN, fp)!=NULL) {
			lineCounter++;
			directive = strtok(buffer,"=");
			value = strtok(NULL," =");
			if (!strncmp(directive, "log_file",8)) {
				new_log_file=malloc(strlen(value));
				if(new_log_file == NULL) {
					snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_logging: Unable to memory for logfile string - %s", strerror(errno));
					write_log(OCTOPUS_LOG_EXIT, log_string);
				}
				new_log_file=value;
				new_log_file[strlen(new_log_file)-1]= '\0';
				break;
			}
		}
		fclose(fp);

		if(new_log_file == NULL) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR! log_file directive not found in conf file");
			write_log(OCTOPUS_LOG_EXIT, log_string);
		}

	}
	else {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR! unable to open conf file! - %s\n", strerror(errno));
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}

	log= fopen(new_log_file, "a");
	if(log == NULL) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_logging: Unable to open log file for writing - %s", strerror(errno));
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}

	group->log_file = log;
	write_log(OCTOPUS_LOG_STD, "STARTUP: logging subsystem initialized");
	return 0;
}


/* statically allocate all the memory required for handling client, member and ghost send/recv buffers */
/* static allocation is memory hungry but it pays off in terms of stability and performance (no malloc required ondemand) */
SESSION* initialize_sessions() {
	int i;
	int supportable_sessions=0;

	//now work out how many sessions we can support with this amount of fds
	supportable_sessions = (group->fd_limit - 4) / 3;
	snprintf(log_string, OCTOPUS_LOG_LEN, "STARTUP: initialize_sessions: With %d fds we can support %d sessions", group->fd_limit, supportable_sessions);
	write_log(OCTOPUS_LOG_STD, log_string);
	if(group->session_limit == 0) {
		group->session_limit = supportable_sessions;
	}
	//if the user has set the session limit higher than the supportable amount then change it and notify via logfile
	else if(group->session_limit > supportable_sessions) {
		group->session_limit = supportable_sessions;
		snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: initialize_sessions: User requested more sessions than we can support, setting session limit to %d", group->session_limit);
		write_log(OCTOPUS_LOG_STD, log_string);
	}
	else {
		snprintf(log_string, OCTOPUS_LOG_LEN, "STARTUP: initialize_sessions: Using user defined session limit of %d sessions", group->session_limit);
		write_log(OCTOPUS_LOG_STD, log_string);
	}

	sessions = malloc(sizeof(SESSION) * group->session_limit);
	if(sessions == NULL) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_sessions: Unable to allocate memory for sessions - %s", strerror(errno));
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}
	//set all the sessions to default (unused) values
	memset(sessions,'\0',(sizeof(SESSION) * group->session_limit));
	for(i=0; i<group->session_limit;i++) {
		sessions[i].state=0;
		sessions[i].clientfd=-1;
		sessions[i].memberfd=-1;
		sessions[i].clonefd=-1;
		sessions[i].client_used_buffer=0;
		sessions[i].member_used_buffer=0;
		sessions[i].clone_used_buffer=0;
	}
	snprintf(log_string, OCTOPUS_LOG_LEN, "STARTUP: initialize_sessions: initialized %d sessions", group->session_limit);
	write_log(OCTOPUS_LOG_STD, log_string);
	return 0;
}
/* try to set file desciptor limits and then initialize FD structure array */
int initialize_fds() {
	int status;
	struct rlimit limit;
	if(group->fd_limit > 0) {
		limit.rlim_max=group->fd_limit;
		limit.rlim_cur=group->fd_limit;
		status=setrlimit(RLIMIT_NOFILE, &limit);
		if(status!=0) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: set_fd_limits: Unable to set fd limit");
		}
		else {
			snprintf(log_string, OCTOPUS_LOG_LEN, "STARTUP: set_fd_limits: Set fd limit to %d", group->fd_limit);
		}
		write_log(OCTOPUS_LOG_STD, log_string);
	}

	status=getrlimit(RLIMIT_NOFILE, &limit);
	if(status==0) {
		group->fd_limit=limit.rlim_cur;
	}
	else {
		group->fd_limit=0;
		snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: set_fd_limits: Unable to get fd limit");
		write_log(OCTOPUS_LOG_STD, log_string);
	}

	fds = malloc(sizeof(FD) * group->fd_limit);
	if(fds == NULL) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_fds: Unable to initialize file descriptor set  - %s", strerror(errno));
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}
	memset(fds,'\0',(sizeof(FD) * group->fd_limit));
	snprintf(log_string, OCTOPUS_LOG_LEN, "STARTUP: initialize_fds: initialized file descriptor set with %d elements", group->fd_limit);
	write_log(OCTOPUS_LOG_STD, log_string);
	return 0;
}

/* set all defaults for group settings */
int initialize_group() {
	int i;
	group = malloc(sizeof(GROUP));
	if(group == NULL) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: Unable to allocate memory for group: %s", strerror(errno));
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}
	memset(group, '\0', sizeof(GROUP));
	group->alive=1;
	group->nmembers=0;
	group->nclones=0;
	group->monitor_interval=DEFAULT_MONITOR_INTERVAL;
	group->connect_timeout=DEFAULT_CONNECT_TIMEOUT;
	group->algorithm=ALGORITHM_LC;
	group->clone_mode=MODE_OFF;
	group->overload_mode=MODE_RELAXED;
	group->shmid=-1;
	group->use_syslog=0;
#ifdef USE_SNMP
	group->snmp_status=SNMP_DISABLED;
	group->overall_load=-1;
#else
	group->snmp_status=SNMP_NOT_INCLUDED;
#endif
	group->hash_rebalance_threshold=DEFAULT_REBALANCE_THRESHOLD;
	group->hash_rebalance_size=DEFAULT_REBALANCE_SIZE;
	group->hash_rebalance_interval=DEFAULT_REBALANCE_INTERVAL;
	group->use_member_outbound_ip=0;
	group->use_clone_outbound_ip=0;
	group->default_maxc=DEFAULT_MAXC;
	group->default_maxl=DEFAULT_MAXL;
	group->session_limit=DEFAULT_SESSION_LIMIT;
	group->fd_limit=DEFAULT_FD_LIMIT;
	group->binding_port=DEFAULT_PORT;
	group->session_weight=atof(DEFAULT_SESSION_WEIGHT);
	group->choose_server=choose_lc_server;
	group->shm_perms=DEFAULT_SHM_PERMS;
	inet_aton(DEFAULT_ADDRESS, &group->binding_ip);

	strncpy(group->version, OCTOPUS_VERSION, OCTOPUS_VERSION_LEN);
	for (i=0; i<HASHTABLESIZE; i++) {
		group->hash_table[i]=HASH_VAL_FREE;
	}
	return 0;
}

/* Create Shared Memory and copy GROUP structure into segment */
int initialize_shm() {
	int status;
	int fd;
	char *data = NULL;
	key_t key;
	int shmid;
	umask(0);
	char temp_name[512];
	if(group->shm_run_dir == NULL) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_shm: No shm_run_dir specified! Check config file");
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}
	snprintf(temp_name, 512, "%s/%s-%s-%s-%d", group->shm_run_dir, "octopus", OCTOPUS_VERSION, inet_ntoa(group->binding_ip), group->binding_port);
	group->shm_run_file = malloc((strlen(temp_name)) * sizeof(char));
	if(group->shm_run_file == NULL) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_shm: Unable to allocate memory for shm_run_file: %s", strerror(errno));
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}
	strncpy(group->shm_run_file, temp_name, strlen(temp_name));
	strncpy(group->shm_run_file_fullname, temp_name, SHM_FILE_FULLNAME_MAX_LENGTH);

	/* the RUN_DIR directory should exist, if not, try to create it */
	status = stat(group->shm_run_dir, &stat_buffer);
	if(!S_ISDIR(stat_buffer.st_mode)) {
		if(mkdir(group->shm_run_dir, 01777) < 0) {
			if(errno != EEXIST) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_shm: Unable to create dir %s: %s", group->shm_run_dir, strerror(errno));
				write_log(OCTOPUS_LOG_EXIT, log_string);
			}
		}
	}
	//we should be creating a new .shm file in RUN_DIR. If the file already exists then there is already an instance running on that ip and port combo
	status = stat(group->shm_run_file, &stat_buffer);
	if(status == -1) {
		if ((fd = open(group->shm_run_file, O_CREAT | O_RDWR, group->shm_perms)) == -1) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_shm: Unable to create file %s: %s", group->shm_run_file, strerror(errno));
			write_log(OCTOPUS_LOG_EXIT, log_string);
		}
		close(fd);
	}
	else {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: Unable to create group. File %s already exists. Maybe a server instance is already running?", group->shm_run_file);
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}

	if ((key = ftok(group->shm_run_file, 'Z')) == -1) {
      	snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_shm: Unable to make SHM key - %s", strerror(errno));
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}
	snprintf(log_string, OCTOPUS_LOG_LEN, "STARTUP: initialize_shm: SHM key= %d",key);
	write_log(OCTOPUS_LOG_STD, log_string);

	if ((shmid = shmget(key, sizeof(GROUP), group->shm_perms | IPC_CREAT )) == -1) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_shm: Unable to create server SHM segment - %s", strerror(errno));
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}

	data = shmat(shmid, (void *) 0, 0);
	if (data == (char *) (-1)) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: initialize_shm: Unable to attach to SHM segment - %s", strerror(errno));
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}
	memcpy(data, group, sizeof(GROUP));
	free(group);
	group=(GROUP *)data;
	group->shmid=shmid;
	return 0;

}

int create_serversocket() {
	int listener;
	int status;
	struct sockaddr_in srvaddr;
	int yes=1;

	bzero(&srvaddr, sizeof(srvaddr));
	srvaddr.sin_family = AF_INET;
	memcpy(&srvaddr.sin_addr, &group->binding_ip, sizeof(struct in_addr));
	srvaddr.sin_port = htons((uint16_t)group->binding_port);

	listener = socket(AF_INET, SOCK_STREAM, 0);
	if(listener<0) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: create_server_socket: Unable to create server socket - %s", strerror(errno));
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}
	status = fcntl(listener, F_SETFL, O_NONBLOCK);
	if(status<0) {
		snprintf(log_string, OCTOPUS_LOG_LEN ,"ERROR: Unable to set server socket to NONBLOCK");
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}
	status = setsockopt(listener, IPPROTO_TCP, TCP_NODELAY,(char *) &yes, (socklen_t)sizeof(yes));
	if(status<0) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: Unable to set TCP_NODELAY on server socket");
		write_log(OCTOPUS_LOG_STD, log_string);
	}
	status = setsockopt(listener, SOL_SOCKET, SO_SNDBUF, &sockbufsize, (socklen_t)sizeof(sockbufsize));
	if(status<0) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: Unable to set SNDBUF on server socket");
		write_log(OCTOPUS_LOG_STD, log_string);
	}
	status = setsockopt(listener, SOL_SOCKET, SO_RCVBUF, &sockbufsize, (socklen_t)sizeof(sockbufsize));
	if(status<0) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: Unable to set RCVBUF on server socket");
		write_log(OCTOPUS_LOG_STD, log_string);
	}
	status = setsockopt(listener, SOL_SOCKET, SO_REUSEADDR, &yes, (socklen_t)sizeof(yes));
	if(status<0) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: Unable to set REUSEADDR on server socket");
		write_log(OCTOPUS_LOG_STD, log_string);
	}
	status = bind(listener, (struct sockaddr *)&srvaddr, (socklen_t)sizeof(srvaddr));
	if(status<0) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: Unable to bind to port. Check nothing else is bound on port %d at ip %s", group->binding_port, inet_ntoa(group->binding_ip));
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}
	status= listen(listener, 1000);
	if(status<0) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: Unable to listen() on server socket");
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}


	return listener;
}
