/*
 * Octopus Load Balancer - Server binary.
 *
 * Copyright 2008 Alistair Reay <alreay1@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 */


#include "octopus.h"
#include "config.c"
#include "init.c"
#include "algorithms.c"
#include "stack.c"
#include "monitor.c"
#include "signals.c"
#include "logging.c"
#include "connect.c"

int usage(char *prog_name) {
	fprintf(stderr, "Octopus Load Balancer version " OCTOPUS_VERSION " date " OCTOPUS_DATE"\n");
	fprintf(stderr, "Copyright 2008 Alistair Reay <alreay1@gmail.com>\n\n");
	fprintf(stderr, "Usage : %s [-hf] [-c config-file]\n", prog_name);
	fprintf(stderr, "	-h 		Print help message\n");
	fprintf(stderr, "	-f		Run server in foreground\n");
	fprintf(stderr, "	-c file		Use given config-file\n");
	fprintf(stderr, "	-d level	Set debugging level\n");
	fprintf(stderr, "\n");
	exit(0);
}

int main(int argc, char *argv[]) {
	int listenerfd;
	int i;
	struct sockaddr_in clientaddr;
	socklen_t size;
	int incomingfd;
	int nfds;
	int status;
	SESSION *temp_session;
	signal(SIGCHLD,signal_handler);
	signal(SIGUSR1,signal_handler);
	signal(SIGALRM,signal_handler);
	signal(SIGTERM,signal_handler);
	signal(SIGINT,signal_handler);

	size = sizeof(clientaddr);

	int param_foreground=FOREGROUND_OFF;
	char *param_conf_file = NULL;
	opterr=0;
	if(argc ==1) {
		usage(argv[0]);
		exit(1);
	}
	while((i = getopt(argc, argv,"d:hc:f")) != -1) {
		switch (i) {
			case 'h':
				usage(argv[0]);
				break;
			case 'c':
				param_conf_file= optarg;
				break;
			case 'f':
				param_foreground=FOREGROUND_ON;
				break;
			case 'd':
				debug_level= atoi(optarg);
				break;
			case '?':
				if(isprint(optopt)) {
					snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: main: Unknown option `-%c'.",optopt);
					write_log(OCTOPUS_LOG_EXIT, log_string);
				}
				else {
					snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: main: Unknown option character `\\x%x'.",optopt);
					write_log(OCTOPUS_LOG_EXIT, log_string);
				}
				return 1;
			default:
				abort();
		}
	}
	for(i = optind; i < argc; i++) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: main: Non-option argument %s", argv[i]);
		write_log(OCTOPUS_LOG_EXIT, log_string);
		exit(1);
	}

	initialize_group();
	/*if we are running the server as a daemon then we don't want to suppress stderr errror messages until after the server has started successfully */
	if(param_foreground == FOREGROUND_OFF) {
		group->foreground=FOREGROUND_INIT;
	}
	else {
		group->foreground=param_foreground;
	}

	initialize_logging(param_conf_file);
	parse_config_file(param_conf_file);
	initialize_shm();
	initialize_monitor(argc, argv);
	initialize_fds();
	initialize_sessions();
	initialize_session_stack();
	initialize_process();
	listenerfd = create_serversocket();

	epfd = epoll_create(group->fd_limit);
	null_ev.events = EPOLLERR | EPOLLHUP ;
	ro_ev.events = EPOLLIN | EPOLLERR | EPOLLHUP;
	wr_ev.events = EPOLLOUT | EPOLLERR | EPOLLHUP;
	rw_ev.events = EPOLLIN | EPOLLOUT | EPOLLERR | EPOLLHUP;
	ro_ev.data.fd = listenerfd;
  	if(epoll_ctl(epfd, EPOLL_CTL_ADD, listenerfd, &ro_ev) < 0) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: main: error adding listener fd to epoll: %s", strerror(errno));
		write_log(OCTOPUS_LOG_EXIT, log_string);
	}
	write_log(OCTOPUS_LOG_STD | OCTOPUS_LOG_SYSLOG, "STARTUP: main: octopus startup completed");
	/* startup has completed, we always run in foreground mode until init has completed then go to daemon mode */
	if(group->foreground == FOREGROUND_INIT) {
		group->foreground=FOREGROUND_OFF;
	}
	while(1) {
		nfds= epoll_wait(epfd, events, (group->session_limit * 3), -1);
		if(nfds <= 0) {
			if(errno==EINTR) {
				continue;
			}
			snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_wait returned error: %s", strerror(errno));
			write_log(OCTOPUS_LOG_STD, log_string);
			continue;
		}
		if(debug_level>5) {
			write_log(OCTOPUS_LOG_STD, "DEBUG: epoll_wait returned.");
			for(i=nfds; i>0; i--) {
				snprintf(log_string, OCTOPUS_LOG_LEN, " - event %d @ fd %i",events[i-1].events,events[i-1].data.u32);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
		}
		for(i=0; i <nfds; i++) {
			if ((events[i].data.fd == listenerfd)) {
				if((events[i].events & EPOLLIN) ) {
					/* incomingfd connection */
					incomingfd = accept(listenerfd, (struct sockaddr *)&clientaddr, &size);
					if (incomingfd < 0) {
						if(errno == EMFILE) {
							snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: cannot accept new connection, file descriptor limit reached!");
							write_log(OCTOPUS_LOG_STD, log_string);
							continue;
						}
						else {
							snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: error accepting new connection: %s",strerror(errno));
							write_log(OCTOPUS_LOG_STD, log_string);
							continue;
						}
					}
					if (fcntl(incomingfd, F_SETFL, O_NONBLOCK) < 0) {
						write_log(OCTOPUS_LOG_STD, "WARNING: error accepting incomingfd connection! cannot set NONBLOCK");
					}
					if (setsockopt(incomingfd, IPPROTO_TCP, TCP_NODELAY, (char *) &yes, sizeof(yes)) < 0) {
						write_log(OCTOPUS_LOG_STD, "WARNING: error accepting incomingfd connection! cannot set TCP_NODELAY");
					}

					fds[incomingfd].session= get_new_session();
					if(fds[incomingfd].session == NULL) {
						shutdown(incomingfd, SHUT_RDWR);
				        close(incomingfd);
						snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: disconnected incomingfd request from host %s, port %d, fd %d as free session was not available", inet_ntoa(clientaddr.sin_addr), ntohs(clientaddr.sin_port), incomingfd);
						write_log(OCTOPUS_LOG_STD, log_string);
					}
					else {
						fds[incomingfd].session->state |= STATE_INUSE;
						fds[incomingfd].session->state |= STATE_CLI_CONNECTED;
						fds[incomingfd].session->state |= STATE_CLI_READ_READY;
						fds[incomingfd].session->state |= STATE_FRESH;
						fds[incomingfd].session->clientfd=incomingfd;
						if(debug_level > 1) {
							snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: connect from host %s, port %d, fd %d", inet_ntoa(clientaddr.sin_addr), ntohs(clientaddr.sin_port), incomingfd);
							write_log(OCTOPUS_LOG_STD, log_string);
						}
						ro_ev.data.fd = incomingfd;
					    if(epoll_ctl(epfd, EPOLL_CTL_ADD, incomingfd, &ro_ev) < 0) {
				    		snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: error adding incomingfd fd to epoll set: %s", strerror(errno));
							write_log(OCTOPUS_LOG_STD, log_string);
					    }
					    /* we will connect the client to a server UNLESS we are using HTTP URI Hashing (because we need to see client's requested URI before we can choose a server */
					    else {
					    	if(group->algorithm != ALGORITHM_HASH) {
					    		if ((fds[incomingfd].session->state & STATE_FRESH) && (!(fds[incomingfd].session->state & STATE_MEM_CONNECTED))) {
					    			status = group->choose_server(fds[incomingfd].session);
					    			if(status == -1) {
							    		snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: rejecting connection attempt due to server selection not returning any servers!");
										write_log(OCTOPUS_LOG_STD, log_string);
					    				fds[incomingfd].session->state |= STATE_DELETE;
					    			}
					    		}
					    	}
					    }
					}
				}
				else  {
					write_log(OCTOPUS_LOG_STD, "WARNING: epoll returned unhandled event for listener fd");
				}
			}
			/* existing connection */
			else {
				incomingfd = events[i].data.fd;
				status= stack_push(fds[incomingfd].session);
				if(status != 0) {
					write_log(OCTOPUS_LOG_STD, "WARNING: stack_push error");
					/*if we can't push onto the stack then we can't track the session and we must kill it. */
					delete_session(fds[incomingfd].session);
				}
				/*client */
				if (incomingfd == fds[incomingfd].session->clientfd) {
					if((events[i].events & EPOLLERR) ) {
						if(debug_level > 0) {
							snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: clientfd %d returned EPOLLERR", incomingfd);
							write_log(OCTOPUS_LOG_STD,log_string);
						}
						fds[incomingfd].session->state &= ~STATE_CLI_CONNECTED;
						continue;
					}
					if((events[i].events & EPOLLHUP) ) {
						if(debug_level > 0) {
							snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: clientfd %d returned EPOLLHUP", incomingfd);
							write_log(OCTOPUS_LOG_STD,log_string);
						}
						fds[incomingfd].session->state &= ~STATE_CLI_CONNECTED;
						continue;
					}
					if((events[i].events & EPOLLIN) ) {
						client_read(incomingfd);
					}
					if(events[i].events & EPOLLOUT) {
						client_write(incomingfd);
					}
				}
				/* member */
				else if (incomingfd == fds[incomingfd].session->memberfd) {
					if((events[i].events & EPOLLERR) ) {
						if(debug_level > 0) {
							snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: memberfd %d returned EPOLLERR", incomingfd);
							write_log(OCTOPUS_LOG_STD, log_string);
						}
						/*if no data has been written to the errored member then try to allocate a new one */
						if (fds[incomingfd].session->state & STATE_FRESH) {
							fds[incomingfd].session->state &= ~STATE_MEM_CONNECTED;
                			rw_ev.data.fd=fds[incomingfd].session->memberfd;
					    	if (epoll_ctl(epfd, EPOLL_CTL_DEL, fds[incomingfd].session->memberfd, &rw_ev) < 0) {
					        	snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: unable to remove dead member's fd from epfd: %s", strerror(errno));
								write_log(OCTOPUS_LOG_STD, log_string);
					        }
							shutdown(fds[incomingfd].session->memberfd, SHUT_RDWR);
							close(fds[incomingfd].session->memberfd);
							fds[incomingfd].session->member->c -= 1;
							fds[incomingfd].session->memberfd = -1;
							if(group->choose_server(fds[incomingfd].session) <0) {
								fds[incomingfd].session->state &= ~STATE_MEM_CONNECTED;
								if(debug_level>0) {
									snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: unable to allocate new member for session with dead member connection");
									write_log(OCTOPUS_LOG_STD, log_string);
								}
							}
						}
						/* otherwise mark the session's member as disconnected */
						else {
							fds[incomingfd].session->state &= ~STATE_MEM_CONNECTED;
						}
						continue;
					}
					if((events[i].events & EPOLLHUP) ) {
						if(debug_level>0) {
							snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: memberfd %d returned EPOLLHUP", incomingfd);
							write_log(OCTOPUS_LOG_STD, log_string);
						}
						fds[incomingfd].session->state &= ~STATE_MEM_CONNECTED;
                		null_ev.data.fd=incomingfd;
						if (epoll_ctl(epfd, EPOLL_CTL_MOD, incomingfd, &null_ev) < 0) {
					        	snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: unable to remove dead sessions fd from epfd: %s", strerror(errno));
								write_log(OCTOPUS_LOG_STD, log_string);
						}
						continue;
					}
					if((events[i].events & EPOLLIN) ) {
						member_read(incomingfd);
					}
					if((events[i].events & EPOLLOUT) ) {
						member_write(incomingfd);
					}
				}
				/*clone */
				else if (incomingfd == fds[incomingfd].session->clonefd)  {
					if((events[i].events & EPOLLERR) ) {
						if(debug_level>0) {
							snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: clonefd %d returned EPOLLERR", incomingfd);
							write_log(OCTOPUS_LOG_STD, log_string);
						}
						fds[incomingfd].session->state &= ~STATE_CLO_CONNECTED;
						shutdown(fds[incomingfd].session->clonefd, SHUT_RDWR);
						close(fds[incomingfd].session->clonefd);
						fds[incomingfd].session->clone->c -= 1;
						fds[incomingfd].session->clonefd = -1;
						continue;
					}
					if((events[i].events & EPOLLHUP) ) {
						if(debug_level>0) {
							snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: clonefd %d returned EPOLLHUP", incomingfd);
							write_log(OCTOPUS_LOG_STD, log_string);
						}
						fds[incomingfd].session->state &= ~STATE_CLO_CONNECTED;
						shutdown(fds[incomingfd].session->clonefd, SHUT_RDWR);
						close(fds[incomingfd].session->clonefd);
						fds[incomingfd].session->clone->c -= 1;
						fds[incomingfd].session->clonefd = -1;
						continue;
					}
					if((events[i].events & EPOLLIN) ) {
						clone_read(incomingfd);
					}
					if((events[i].events & EPOLLOUT) ) {
						clone_write(incomingfd);
					}

				}
				else {
					snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: Unhandled socket setting @ fd %d",incomingfd);
					write_log(OCTOPUS_LOG_STD, log_string);

				}
			}
		}
		for(i=0; i<= session_stack.top;)  {
			temp_session=stack_pop();
			if(temp_session == NULL) {
				write_log(OCTOPUS_LOG_STD, "WARNING: Attempt to pop from stack failed");
			}
			else {
				maintain_session_sockets(temp_session);
			}
		}

	}
	return 0;
}
int maintain_session_sockets(SESSION *s) {
	int state;
	state = s->state;
	if(debug_level > 2) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: state = %d for client %d for server %d for clone %d",state, s->clientfd, s->memberfd, s->clonefd);
		write_log(OCTOPUS_LOG_STD, log_string);
	}
	if (state==0) {
		delete_session(s);
		return -1;
	}
	if (state & STATE_DELETE) {
		delete_session(s);
		return 0;
	}
	if(state & STATE_CLO_CONNECTED) {
		if((state & STATE_CLO_WRITE_READY) && (state & STATE_CLO_READ_READY)) {
			if(debug_level > 2) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: setting clone for r/w @ fd %d",s->clonefd);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
			rw_ev.data.fd=s->clonefd;
			if(epoll_ctl(epfd, EPOLL_CTL_MOD, s->clonefd, &rw_ev) <0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_mod failed to set clone for r/w");
				write_log(OCTOPUS_LOG_STD, log_string);
			}

		}
		else if(state & STATE_CLO_WRITE_READY) {
			if(debug_level > 2) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: setting clone for write @ fd %d",s->clonefd);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
			wr_ev.data.fd=s->clonefd;
			if(epoll_ctl(epfd, EPOLL_CTL_MOD, s->clonefd, &wr_ev) < 0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_mod failed to set clone for write");
				write_log(OCTOPUS_LOG_STD, log_string);
			}
		}
		else if(state & STATE_CLO_READ_READY) {
			if(debug_level > 2) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: setting clone for read @ fd %d",s->clonefd);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
			ro_ev.data.fd=s->clonefd;
			if(epoll_ctl(epfd, EPOLL_CTL_MOD, s->clonefd, &ro_ev) <0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_mod failed to set clone for read");
				write_log(OCTOPUS_LOG_STD, log_string);
			}
		}
		else {
			if(debug_level > 2) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: setting clone for null @ fd %d",s->clonefd);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
			null_ev.data.fd=s->clonefd;
			if(epoll_ctl(epfd, EPOLL_CTL_MOD, s->clonefd, &null_ev) <0) {
					snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_mod failed to set clone for null");
					write_log(OCTOPUS_LOG_STD, log_string);
			}

		}
	}

	if(state & STATE_CLI_CONNECTED) {
		if((state & STATE_CLI_WRITE_READY) && (state & STATE_CLI_READ_READY)) {
			if(debug_level > 2) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: setting client for r/w @ fd %d",s->clientfd);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
			rw_ev.data.fd=s->clientfd;
			if(epoll_ctl(epfd, EPOLL_CTL_MOD, s->clientfd, &rw_ev) <0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_mod failed to set client for r/w");
				write_log(OCTOPUS_LOG_STD, log_string);
			}
		}
		else if(state & STATE_CLI_WRITE_READY) {
			if(debug_level > 2) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: setting client for write @ fd %d",s->clientfd);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
			wr_ev.data.fd=s->clientfd;
			if(epoll_ctl(epfd, EPOLL_CTL_MOD, s->clientfd, &wr_ev) <0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_mod failed to set client for write");
				write_log(OCTOPUS_LOG_STD, log_string);
			}
		}
		else if(state & STATE_CLI_READ_READY) {
			if(debug_level > 2) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: setting client for read @ fd %d",s->clientfd);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
			ro_ev.data.fd=s->clientfd;
			if(epoll_ctl(epfd, EPOLL_CTL_MOD, s->clientfd, &ro_ev) <0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_mod failed to set client for read");
				write_log(OCTOPUS_LOG_STD, log_string);
			}
		}
		else {
			if(debug_level > 2) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: setting client for null @ fd %d",s->clientfd);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
			null_ev.data.fd=s->clientfd;
			if(epoll_ctl(epfd, EPOLL_CTL_MOD, s->clientfd, &null_ev) <0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_mod failed to set client for null");
				write_log(OCTOPUS_LOG_STD, log_string);
			}

		}
	}
	else {
		if(debug_level > 2) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: deleting session as state shows client disconnected");
			write_log(OCTOPUS_LOG_STD, log_string);
		}
		delete_session(s);
		return 0;
	}
	/* member */
	if(state & STATE_MEM_CONNECTED) {
		if((state & STATE_MEM_WRITE_READY) && (state & STATE_MEM_READ_READY)) {
			if(debug_level > 2) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: setting member for r/w @ fd %d",s->memberfd);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
			rw_ev.data.fd=s->memberfd;
			if(epoll_ctl(epfd, EPOLL_CTL_MOD, s->memberfd, &rw_ev) <0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_mod failed to set member for r/w");
				write_log(OCTOPUS_LOG_STD, log_string);
			}
		}
		else if(state & STATE_MEM_WRITE_READY) {
			if(debug_level > 2) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: setting member for write @ fd %d",s->memberfd);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
			wr_ev.data.fd=s->memberfd;
			if(epoll_ctl(epfd, EPOLL_CTL_MOD, s->memberfd, &wr_ev) < 0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_mod failed to set member for write");
				write_log(OCTOPUS_LOG_STD, log_string);
			}
		}
		else if(state & STATE_MEM_READ_READY) {
			if(debug_level > 2) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: setting member for read @ fd %d",s->memberfd);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
			ro_ev.data.fd=s->memberfd;
			if(epoll_ctl(epfd, EPOLL_CTL_MOD, s->memberfd, &ro_ev) <0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_mod failed to set member for read");
				write_log(OCTOPUS_LOG_STD, log_string);
			}
		}
		else {
			if(debug_level > 2) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: setting member for null @ fd %d",s->memberfd);
				write_log(OCTOPUS_LOG_STD, log_string);
			}
			null_ev.data.fd=s->memberfd;
			if(epoll_ctl(epfd, EPOLL_CTL_MOD, s->memberfd, &null_ev) <0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: epoll_mod failed to set member for null");
				write_log(OCTOPUS_LOG_STD, log_string);
			}
		}
	}
	else {
		/* a situation will exist where a member server has disconnected (EOF, error etc) but there is still more data to be written to the client */
		if ((state & STATE_CLI_WRITE_READY) ==0) {
			/*if we are in strict mode and clone is still connected then we don't kill the session */
			if((group->clone_mode==MODE_STRICT) && (state & STATE_CLO_CONNECTED)) {
				if(debug_level > 3) {
					snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: not deleting session as clone server not finished");
					write_log(OCTOPUS_LOG_STD, log_string);
				}
				return 0;
			}
			else {
				if(debug_level > 2) {
					snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: deleting as state shows server disconnected and no more data for client");
					write_log(OCTOPUS_LOG_STD, log_string);
				}
				delete_session(s);
				return 0;
			}
		}
	}
	return 0;

}
int client_read(int fd) {
	errno=0;
	int status;
	nbytes= read(fd, (fds[fd].session->client_read_buffer + fds[fd].session->client_used_buffer), (MESSAGE_SIZE_LIMIT - fds[fd].session->client_used_buffer));
	if (nbytes <= 0) {
		if (nbytes == 0) {
			/* EOF */
			if(debug_level > 3) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: client_read: read() returned EOF");
				write_log(OCTOPUS_LOG_STD,log_string);
			}
			fds[fd].session->state &= ~STATE_CLI_CONNECTED;
			return -1;
		}
		/* socket error! */
		if(debug_level > 3) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: client_read: read() returned an error: %s",strerror(errno));
			write_log(OCTOPUS_LOG_STD,log_string);
		}
		fds[fd].session->state &= ~STATE_CLI_CONNECTED;
		return -1;
	}
	fds[fd].session->client_used_buffer += (int)nbytes;
	/* this check checks if the session has established a connection to a server and if not, connects it */
	if ((fds[fd].session->state & STATE_FRESH) && (!(fds[fd].session->state & STATE_MEM_CONNECTED))) {
		status = group->choose_server(fds[fd].session);
		if(status == -1) {
			fds[fd].session->state |= STATE_DELETE;
			return -1;
		}
	}

	if((fds[fd].session->state & STATE_CLO_CONNECTED) >0) {
		if(nbytes > (MESSAGE_SIZE_LIMIT - fds[fd].session->clone_used_buffer)) {
			/* in the situation where the clone has not been able to be written to enough, we have to cut it loose to avoid any ugliness */
			fds[fd].session->state &= ~STATE_CLO_CONNECTED;
            null_ev.data.fd=fds[fd].session->clonefd;
			if (epoll_ctl(epfd, EPOLL_CTL_MOD, fds[fd].session->clonefd, &null_ev) < 0) {
			   	snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: unable to remove dead clone's fd from epfd: %s", strerror(errno));
				write_log(OCTOPUS_LOG_STD, log_string);
			}
		}
		else {
			strncpy(fds[fd].session->clone_write_buffer + fds[fd].session->clone_used_buffer, (fds[fd].session->client_read_buffer + fds[fd].session->client_used_buffer - nbytes), nbytes);
			fds[fd].session->clone_used_buffer += nbytes;
		}
	}
	if(debug_level > 2) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: client_read: read %d bytes from client @ fd %d",(int)nbytes,fd);
		write_log(OCTOPUS_LOG_STD, log_string);
	}

	/* bytes were read from the client */
	fds[fd].session->state |= STATE_MEM_WRITE_READY;
	fds[fd].session->state |= STATE_CLO_WRITE_READY;

	/* client input buffer is full */
	if(fds[fd].session->client_used_buffer == MESSAGE_SIZE_LIMIT) {
		fds[fd].session->state |= STATE_CLI_BUFF_FULL;
		fds[fd].session->state &= ~STATE_CLI_READ_READY;

	}
	return 0;
}

int client_write(int fd) {
	nbytes = write(fd, fds[fd].session->member_read_buffer, fds[fd].session->member_used_buffer);
	if(debug_level > 2) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: client_write: wrote %d bytes to client @ fd %d",(int)nbytes,fd);
		write_log(OCTOPUS_LOG_STD, log_string);
	}

	if (nbytes <= 0) {
		if(debug_level > 1) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "ERROR: client_write: write() failed: %s",strerror(errno));
			write_log(OCTOPUS_LOG_STD,log_string);
		}
		fds[fd].session->state ^= STATE_CLI_CONNECTED;
		return -1;
	}
	fds[fd].session->member_used_buffer -= (int)nbytes;

	/*no more data for client */
	if(fds[fd].session->member_used_buffer == 0) {
		fds[fd].session->state &= ~STATE_CLI_WRITE_READY;
	}
	/*there is still more to be written to the client! move any unwritten data to the front of the buffer */
	else {
		memcpy(fds[fd].session->member_read_buffer, (fds[fd].session->member_read_buffer + nbytes), fds[fd].session->member_used_buffer);
	}
	/*we just wrote some stuff to the client, the member buffer will have free space */
	if (fds[fd].session->state & STATE_MEM_BUFF_FULL) {
		fds[fd].session->state &= ~STATE_MEM_BUFF_FULL;
		if(fds[fd].session->state & STATE_MEM_CONNECTED) {
			fds[fd].session->state |= STATE_MEM_READ_READY;
		}
	}
	return 0;
}

/* member is ready for writing */
int member_write(int fd) {
	nbytes = write(fd, fds[fd].session->client_read_buffer, fds[fd].session->client_used_buffer);
	if (nbytes <= 0) {
		if(debug_level > 3) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: member_write: ERROR, unsetting STATE_MEM_CONN for fd @ %d",fd);
			write_log(OCTOPUS_LOG_STD,log_string);
		}
		fds[fd].session->state &= ~STATE_MEM_CONNECTED;
		return -1;
	}
	if(debug_level > 2) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: member_write: wrote %d bytes to server @ fd %d",(int)nbytes,fd);
		write_log(OCTOPUS_LOG_STD, log_string);
	}
	if(fds[fd].session->state & STATE_FRESH ) {
		fds[fd].session->state &= ~STATE_FRESH;
	}
	fds[fd].session->client_used_buffer -= (int)nbytes;
	fds[fd].session->member->bsent += (int)nbytes;
	fds[fd].session->state |= STATE_MEM_READ_READY;
	/*if the buffer is now empty, then we don't need to montitor server for more writes */
	if(fds[fd].session->client_used_buffer == 0) {
		if(debug_level > 3) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: member_write: server buffer empty. Unsetting STATE_MEM_WRITE_READY for fd @ %d",fd);
			write_log(OCTOPUS_LOG_STD,log_string);
		}
		fds[fd].session->state &= ~STATE_MEM_WRITE_READY;
	}
	/*if the buffer was not fully emptied, move everything that wasn't written to the front of the buffer */
	else {
		memcpy(fds[fd].session->client_read_buffer, (fds[fd].session->client_read_buffer + nbytes), fds[fd].session->client_used_buffer);
	}
	/*if the client buffer _was_ full, then unset the condition and reenq the client */
	if(fds[fd].session->state & STATE_CLI_BUFF_FULL) {
		fds[fd].session->state &= ~STATE_CLI_BUFF_FULL;
		fds[fd].session->state |= STATE_CLI_READ_READY;
	}
	return 0;
}




int member_read(int fd) {
	nbytes= read(fd, ((fds[fd].session->member_read_buffer) + fds[fd].session->member_used_buffer), (MESSAGE_SIZE_LIMIT - fds[fd].session->member_used_buffer));
	if (nbytes <= 0) {
		if(nbytes < 0) {
			if(debug_level > 3) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: member_read: ERROR, unsetting STATE_MEM_CONNECTED for fd @ %d",fd);
				write_log(OCTOPUS_LOG_STD,log_string);
			}
			fds[fd].session->state ^= STATE_MEM_CONNECTED;
			return -1;
		}
		if(debug_level > 3) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: member_read: EOF, unsetting STATE_MEM_CONNECTED for fd @ %d",fd);
			write_log(OCTOPUS_LOG_STD,log_string);
		}
		fds[fd].session->state ^= STATE_MEM_CONNECTED;
		fds[fd].session->member->c -=1;
		fds[fd].session->member->completed_c ++;
        null_ev.data.fd=fds[fd].session->memberfd;
        if (epoll_ctl(epfd, EPOLL_CTL_DEL, fds[fd].session->memberfd, &null_ev) < 0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: member_read: epoll_mod failed to delete memberfd");
				write_log(OCTOPUS_LOG_STD, log_string);
        }
		shutdown(fds[fd].session->memberfd, SHUT_RDWR);
		close(fds[fd].session->memberfd);
		fds[fd].session->memberfd=-1;
		return 0;
	}

	if(debug_level > 2) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: member_read: read %d bytes from server @ fd %d",(int)nbytes,fd);
		write_log(OCTOPUS_LOG_STD, log_string);
	}
	fds[fd].session->member_used_buffer += (int)nbytes;
	fds[fd].session->member->brecv +=nbytes;
	/* reading data from a member will always mean we have to write to the client */
	fds[fd].session->state |= STATE_CLI_WRITE_READY;

	if(fds[fd].session->member_used_buffer == MESSAGE_SIZE_LIMIT) {
		if(debug_level > 3) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: member_read: server buffer full. Unsetting SRV_READ_READY for fd @ %d",fd);
			write_log(OCTOPUS_LOG_STD,log_string);
		}
        fds[fd].session->state |= STATE_MEM_BUFF_FULL;
        fds[fd].session->state &= ~STATE_MEM_READ_READY;
	}
	return 0;
}

int clone_write(int fd) {
	nbytes =write(fd,fds[fd].session->clone_write_buffer, fds[fd].session->clone_used_buffer);
	if (nbytes <=0) {
		if(debug_level > 3) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: clone_write: ERROR, unsetting STATE_CLO_CONN for fd @ %d",fd);
			write_log(OCTOPUS_LOG_STD,log_string);
		}
		fds[fd].session->state &= ~STATE_CLO_CONNECTED;
		null_ev.data.fd=fd;
		if (epoll_ctl(epfd, EPOLL_CTL_MOD, fd, &null_ev) < 0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: clone_write: epoll_mod failed to set clone for null");
				write_log(OCTOPUS_LOG_STD, log_string);
		}
		return 0;
	}
	if(debug_level > 2) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: clone_write: wrote %d bytes to clone @ fd %d",(int)nbytes,fd);
		write_log(OCTOPUS_LOG_STD, log_string);
	}
	fds[fd].session->clone_used_buffer -= (int)nbytes;
	fds[fd].session->clone->bsent +=nbytes;
	/*after a write, we expect a response */
	fds[fd].session->state |= STATE_CLO_READ_READY;

	if(fds[fd].session->clone_used_buffer==0) {
		if(debug_level > 3) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: clone_write: clone buffer empty. Unsetting STATE_CLO_WRITE_READY for fd @ %d",fd);
			write_log(OCTOPUS_LOG_STD,log_string);
		}
		fds[fd].session->state &= ~STATE_CLO_WRITE_READY;
	}
	/*if the buffer was not fully emptied, move everything that wasn't written to the front of the buffer */
	else {
		memcpy(fds[fd].session->clone_write_buffer, (fds[fd].session->clone_write_buffer + nbytes), fds[fd].session->clone_used_buffer);
	}
	return 0;
}

int clone_read(int fd) {
	nbytes= read(fd, waste_buffer, MESSAGE_SIZE_LIMIT);
	if (nbytes <= 0) {
		if(nbytes < 0) {
			if(debug_level > 3) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: clone_read: ERROR, unsetting STATE_CLO_CONNECTED for fd @ %d",fd);
				write_log(OCTOPUS_LOG_STD,log_string);
			}
			fds[fd].session->state ^= STATE_CLO_CONNECTED;
			return -1;
		}
		if(debug_level > 3) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: clone_read: EOF, unsetting STATE_CLO_CONNECTED for fd @ %d",fd);
			write_log(OCTOPUS_LOG_STD,log_string);
		}
		fds[fd].session->state ^= STATE_CLO_CONNECTED;
		fds[fd].session->clone->c -=1;
		fds[fd].session->clone->completed_c ++;
        null_ev.data.fd=fds[fd].session->clonefd;
        if (epoll_ctl(epfd, EPOLL_CTL_DEL, fds[fd].session->clonefd, &null_ev) < 0) {
				snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: clone_read: epoll_mod failed to delete clonefd");
				write_log(OCTOPUS_LOG_STD, log_string);
        }
		shutdown(fds[fd].session->clonefd, SHUT_RDWR);
		close(fds[fd].session->clonefd);
		fds[fd].session->clonefd=-1;
		return 0;
	}
	if(debug_level > 2) {
		snprintf(log_string, OCTOPUS_LOG_LEN, "DEBUG: clone_read: read %d bytes from clone @ fd %d",(int)nbytes,fd);
		write_log(OCTOPUS_LOG_STD, log_string);
	}
	fds[fd].session->clone->brecv +=nbytes;
	return 0;
}

SESSION* get_new_session() {
	int i;
	for(i=0; i< group->session_limit; i++) {
		if ((sessions[i].state) == STATE_UNUSED) {
			return &sessions[i];
		}
	}
	write_log(OCTOPUS_LOG_STD, "WARNING: get_new_session: unable to supply new session!!");
	return NULL;
}

int delete_session(SESSION *session) {
	if(session->clientfd >= 0) {
		null_ev.data.fd=session->clientfd;
		if (epoll_ctl(epfd, EPOLL_CTL_DEL, session->clientfd, &null_ev) < 0) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: delete_session: epoll_mod failed to delete fd");
			write_log(OCTOPUS_LOG_STD, log_string);
		}
		shutdown(session->clientfd, SHUT_RDWR);
		close(session->clientfd);
	}
	if(session->memberfd >= 0) {
		session->member->c -=1;
		session->member->completed_c ++;
        null_ev.data.fd=session->memberfd;
        if (epoll_ctl(epfd, EPOLL_CTL_DEL, session->memberfd, &null_ev) < 0) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: delete_session: epoll_mod failed to delete fd");
			write_log(OCTOPUS_LOG_STD, log_string);
        }
		shutdown(session->memberfd, SHUT_RDWR);
		close(session->memberfd);
	}
	if(session->clonefd >= 0) {
		session->clone->c -=1;
		session->clone->completed_c ++;
        null_ev.data.fd=session->clonefd;
        if (epoll_ctl(epfd, EPOLL_CTL_DEL, session->clonefd, &null_ev) < 0) {
			snprintf(log_string, OCTOPUS_LOG_LEN, "WARNING: delete_session: epoll_mod failed to delete fd");
			write_log(OCTOPUS_LOG_STD, log_string);
        }
		shutdown(session->clonefd, SHUT_RDWR);
		close(session->clonefd);
	}

	session->state=0;
	session->clientfd=-1;
	session->memberfd=-1;
	session->clonefd=-1;
	session->client_used_buffer=0;
	session->member_used_buffer=0;
	session->clone_used_buffer=0;

	return 0;
}

