/*
 * Octopus Load Balancer - Main header file.
 *
 * Copyright 2008 Alistair Reay <alreay1@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <sysexits.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/shm.h>
#include <sys/ipc.h>
#include <sys/epoll.h>
#include <sys/times.h>
#include <sys/syslog.h>
#include <dirent.h>
#ifdef USE_SNMP
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/session_api.h>
#endif

#define OCTOPUS_VERSION "1.06"
#define OCTOPUS_DATE "02/06/2009"
#define OCTOPUS_VERSION_LEN 16

#define FOREGROUND_OFF 0
#define FOREGROUND_ON 1
#define FOREGROUND_INIT 2

#define DEFAULT_PORT 1080
#define DEFAULT_ADDRESS "0.0.0.0"
#define DEFAULT_SESSION_WEIGHT "0.1"
#define DEFAULT_MONITOR_INTERVAL 5
#define DEFAULT_CONNECT_TIMEOUT 5

#define SNMP_NOT_INCLUDED 0
#define SNMP_DISABLED 1
#define SNMP_ENABLED 2

#define CLIENT_MAX_INPUT 128

#define MAXBALANCERS 256
#define DEFAULT_MAXC 400
#define DEFAULT_MAXL 4.0

#define OCTOPUS_LOG_STD 0
#define OCTOPUS_LOG_EXIT 1
#define OCTOPUS_LOG_SYSLOG 2
#define OCTOPUS_LOG_LEN 512

#define SERVER_STATE_FREE -3
#define SERVER_STATE_DELETED -2
#define SERVER_STATE_FAILED -1
#define SERVER_STATE_DISABLED 0
#define SERVER_STATE_ENABLED 1

#define MAXSERVERS 512
#define MESSAGE_SIZE_LIMIT 4096
#define HASHTABLESIZE 65535

#define DEFAULT_SHM_PERMS 0644

/* zero means auto configure based on available fds */
#define DEFAULT_SESSION_LIMIT 0
#define DEFAULT_FD_LIMIT 0

/* session states */
#define STATE_UNUSED 0
#define STATE_INUSE 1
#define STATE_MEM_CONNECTED 2
#define STATE_CLI_CONNECTED 4
#define STATE_CLO_CONNECTED 8
#define STATE_MEM_READ_READY 16
#define STATE_CLI_READ_READY 32
#define STATE_CLO_READ_READY 64
#define STATE_MEM_WRITE_READY 128
#define STATE_CLI_WRITE_READY 256
#define STATE_CLO_WRITE_READY 512
#define STATE_MEM_BUFF_FULL 1024
#define STATE_CLI_BUFF_FULL 2048
#define STATE_FRESH 4096
#define STATE_DELETE 8192

#define DEFAULT_SHM_RUN_DIR "/var/run/octopus/"

#define HASH_VAL_FREE -1

#define DEFAULT_REBALANCE_THRESHOLD 30
#define DEFAULT_REBALANCE_SIZE 5
#define DEFAULT_REBALANCE_INTERVAL 30

#define SHM_FILE_FULLNAME_MAX_LENGTH 2048
/* modes used for both cloning and overload */
#define MODE_OFF 0
#define MODE_FAST 1
#define MODE_RELAXED 1
#define MODE_STRICT 2

#define ALGORITHM_RR 1
#define ALGORITHM_LC 2
#define ALGORITHM_LL 3
#define ALGORITHM_HASH 4

#define SERVERNAME_MAX_LENGTH 16

#define NO_USE_CLONE 0
#define USE_CLONE 1

#define MAX_CONF_LINE_LEN 100

#define URI_LINE_LEN 200

typedef struct {
	unsigned short int id;
	char name[SERVERNAME_MAX_LENGTH];	/* servername */
	int status;	/* enabled, disabled or failed */
	int port;	/* which port do we connect to for this memeber? */
	struct sockaddr_in myaddr;	/* socket for server */
	int c;	/* current number of connections */
	int maxc;	/* maximum numbre of connections */
	unsigned long int completed_c;	/* completed number of connections */
	unsigned long bsent;	/* bytes sent */
	unsigned long brecv;	/* bytes received */
	float load;	/* snmp reported load */
	float maxload;	/* user specified max load */
	int e_load; /* effective loading, maxload/load * 100 */
	unsigned short int hash_table_usage;
} SERVER;

typedef struct {
	unsigned int state;
	int clientfd;
	int memberfd;
	int clonefd;
	char member_read_buffer[MESSAGE_SIZE_LIMIT];
	char client_read_buffer[MESSAGE_SIZE_LIMIT];
	char clone_write_buffer[MESSAGE_SIZE_LIMIT];
	int member_used_buffer;
	int client_used_buffer;
	int clone_used_buffer;
	SERVER *member;
	SERVER *clone;
} SESSION;

typedef struct {
	SESSION *session;
} FD;


typedef struct {
	short int hash_table[HASHTABLESIZE];
	unsigned short int nmembers;
	unsigned short int nclones;
	unsigned short int monitor_interval;
	int alive;
	int algorithm;
	int clone_mode;
	int overload_mode;
	SERVER members[MAXSERVERS];
	SERVER clones[MAXSERVERS];
	int (*choose_server)(SESSION *session);
	u_char *snmp_community_pw;
	int snmp_status;
	int binding_port;
	struct in_addr binding_ip;
	struct in_addr member_outbound_ip;
	struct in_addr clone_outbound_ip;
	int connect_timeout;
	pid_t monitor_pid;
	pid_t master_pid;
	int shmid;
	int foreground;
	FILE *log_file;
	float session_weight;
	int hash_rebalance_threshold; /* if any two servers have e_load difference of this percentage then move some hashes to lowest loaded server */
	int hash_rebalance_size; /* move this percentage of currently assigned hashes per go */
	int hash_rebalance_interval; /* do rebalancing every X seconds */
	int use_member_outbound_ip;
	int use_clone_outbound_ip;
	char *shm_run_dir;
	char *shm_run_file;
	char shm_run_file_fullname[SHM_FILE_FULLNAME_MAX_LENGTH];
	int default_maxc;
	float default_maxl;
	char version[OCTOPUS_VERSION_LEN];
	int fd_limit;
	int session_limit;
	int use_syslog;
	int shm_perms;
	int overall_load;
} GROUP;

typedef struct {
        SESSION **contents;
        int top;
        int maxsize;
}stackT;

/* function prototypes */
int handle_connection(int client_fd, char *buffer, size_t buffer_size);
int create_group_server(int clone_server, char *serverName, int serverStatus, int serverPort, struct in_addr *serverIP, int serverMaxc, float serverMaxload);
SESSION* get_new_session();
int member_read(int fd);
int member_write(int fd);
int client_read(int fd);
int client_write(int fd);
int clone_read(int fd);
int clone_write(int fd);
int delete_session(SESSION *session);
int choose_hash_server(SESSION *session);
int choose_rr_server(SESSION *session);
int choose_ll_server(SESSION *session);
int choose_lc_server(SESSION *session);
int set_lc_server();
int set_ll_server();
int set_rr_server();
int set_hash_server(SESSION *session);
int connect_server(SESSION *session, int use_clone);
int maintain_session_sockets();
int calc_effective_load();
int connect_to_shm(char *run_file, int ignore_version_check);
int rebalance_hash();
int clean_exit();
int write_log(int options, char *description);

/* some global variables */
GROUP *group;
SESSION *sessions;
FD *fds;
int epfd;
struct epoll_event null_ev;
struct epoll_event ro_ev;
struct epoll_event wr_ev;
struct epoll_event rw_ev;
struct epoll_event events[20000];
unsigned short int next_member=0;
unsigned short int next_clone=0;
static int sockbufsize = MESSAGE_SIZE_LIMIT;
char uri_delim[] = "?";
ssize_t nbytes;
int yes=1;
stackT session_stack;
struct stat stat_buffer;
char waste_buffer[MESSAGE_SIZE_LIMIT];
int debug_level=0;
char *server_status[8] = {"Free", "Deleted", "Failed", "Disabled", "Enabled"};
char *cloning_status[8] = {"FAILED", "FAILED", "OFF", "FAST", "STRICT"};
char *overload_status[8] = {"RELAXED", "STRICT"};
char *algorithm_status[16] = {"ROUND ROBIN", "LEAST CONNECTIONS", "LEAST LOAD", "HASH"};
char log_string[OCTOPUS_LOG_LEN];

/* this function is used by client and monitor where changing server states may change the cloning status */
int set_cloned_state() {
	int temp=0;
	int i;
	if(group->clone_mode >= 1) {
		for (i=0; i< group->nclones; i++) {
			if(group->clones[i].status <= 0) {
				temp++;
			}
		}
		if(temp == group->nclones) {
			if(group->clone_mode == 1) {
				group->clone_mode= -1;
			}
			else if(group->clone_mode == 2) {
				group->clone_mode= -2;
			}

		}

	}
	else if(group->clone_mode <= -1) {
		for (i=0; i< group->nclones; i++) {
			if(group->clones[i].status == 1) {
				if(group->clone_mode == -1) {
					group->clone_mode= 1;
				}
				else if(group->clone_mode == -2) {
					group->clone_mode= 2;
				}
				break;
			}
		}
	}
	return 0;
}
/*
tries to create a new server
this function is used by the client too so can't write to log directly from here, have to get client and server to do stuff according to ret val
returns 0 on success
returns -1 when malloc fails
returns -2 when tcp port is invalid
returns -3 when IP address is invalid
*/
int create_group_server(int clone_server, char *serverName, int serverStatus, int serverPort, struct in_addr *serverIP, int serverMaxc, float serverMaxload) {
	int i;
	SERVER *s;
	s = malloc(sizeof(SERVER));
	if(s == NULL) {
		return -1;
	}
	memset(s,'\0', sizeof(SERVER));
	strncpy(s->name, serverName, SERVERNAME_MAX_LENGTH);
	s->status=serverStatus;
	s->port=serverPort;
	s->maxc=serverMaxc;
	s->load=-1;
	s->maxload=serverMaxload;
	s->bsent=0;
	s->brecv=0;
	s->e_load=0;
	s->hash_table_usage=0;
	s->myaddr.sin_family = AF_INET;
	if(htons((uint16_t)(serverPort)) > 0) {
		s->myaddr.sin_port = htons((uint16_t)(serverPort));
	}
	else {
		free(s);
		return -2;
	}
	memcpy(&s->myaddr.sin_addr, serverIP, sizeof(struct in_addr));

	/*for a member */
	if(clone_server==0) {
		/*look for free slots first */
		for(i=0; i < group->nmembers; i++) {
			if(group->members[i].status == SERVER_STATE_FREE) {
				memcpy((void *)&group->members[i], (void *)s, sizeof(SERVER));
				group->members[i].id=i;
				free(s);
				return 0;
			}
		}
		/*otherwise just copy to end of members array and increment count */
		memcpy((void *)&group->members[(group->nmembers)], (void *)s, sizeof(SERVER));
		group->members[group->nmembers].id=group->nmembers;
		group->nmembers++;

	}
	/*for a clone */
	else {
		/*look for free slots first */
		for(i=0; i < group->nclones; i++) {
			if(group->clones[i].status == SERVER_STATE_FREE) {
				memcpy((void *)&group->clones[i], (void *)s, sizeof(SERVER));
				group->clones[i].id=i;
				free(s);
				return 0;
			}
		}
		/*otherwise just copy to end of clones array and increment count */
		memcpy((void *)&group->clones[(group->nclones)], (void *)s, sizeof(SERVER));
		group->clones[group->nclones].id=group->nclones;
		group->nclones++;
	}

	free(s);
	return 0;
}

